# encoding: utf-8

"""Unit test suite for `cr.cube.cube` module."""

from __future__ import absolute_import, division, print_function, unicode_literals

import pytest
import numpy as np

from cr.cube.cube import Cube, CubeSet, _Measures
from cr.cube.cubepart import _Slice, _Strand, _Nub
from cr.cube.enums import DIMENSION_TYPE as DT
from cr.cube.dimension import Dimension

from ..fixtures import CR  # ---mnemonic: CR = 'cube-response'---
from ..unitutil import call, class_mock, instance_mock, property_mock, method_mock


class DescribeCubeSet(object):
    """Unit-test suite for `cr.cube.cube.CubeSet` object."""

    @pytest.mark.parametrize(
        ("cubes_dimtypes", "expected_value"),
        (
            ((), False),
            (((DT.CAT, DT.CAT),), False),
            (((DT.TEXT,), (DT.TEXT,)), False),
            (((DT.CAT,), (DT.CAT, DT.CAT)), True),
            (((DT.TEXT,), (DT.CAT, DT.CAT)), True),
            (((DT.MR_CAT,), (DT.CAT, DT.CAT)), True),
            (((DT.CAT,), (DT.CAT, DT.MR_CAT)), False),
            (((DT.CAT, DT.CAT), (DT.CAT, DT.MR_CAT)), False),
            (((DT.DATETIME,), (DT.DATETIME, DT.DATETIME)), True),
            (((DT.DATETIME, DT.DATETIME), (DT.DATETIME, DT.DATETIME)), True),
            (((DT.DATETIME, DT.DATETIME), (DT.DATETIME, DT.MR_CAT)), False),
        ),
    )
    def it_knows_whether_it_can_show_pairwise(
        self, request, cubes_dimtypes, expected_value, _cubes_prop_
    ):
        _cubes_prop_.return_value = tuple(
            instance_mock(
                request, Cube, dimension_types=cube_dimtypes, ndim=len(cube_dimtypes)
            )
            for cube_dimtypes in cubes_dimtypes
        )
        cube_set = CubeSet(None, None, None, None)

        can_show_pairwise = cube_set.can_show_pairwise

        assert can_show_pairwise is expected_value

    def it_knows_its_description(self, _cubes_prop_, cube_):
        cube_.description = "Are you male or female?"
        _cubes_prop_.return_value = (cube_,)
        cube_set = CubeSet(None, None, None, None)

        description = cube_set.description

        assert description == "Are you male or female?"

    @pytest.mark.parametrize(
        ("first_cube_has_means", "expected_value"), ((True, True), (False, False))
    )
    def it_knows_whether_it_has_means(
        self, first_cube_has_means, expected_value, _cubes_prop_, cube_
    ):
        cube_.has_means = first_cube_has_means
        _cubes_prop_.return_value = (cube_,)
        cube_set = CubeSet(None, None, None, None)

        has_means = cube_set.has_means

        assert has_means == expected_value

    @pytest.mark.parametrize(
        ("first_cube_has_w_counts", "expected_value"), ((True, True), (False, False))
    )
    def it_knows_whether_it_has_weighted_counts(
        self, first_cube_has_w_counts, expected_value, _cubes_prop_, cube_
    ):
        cube_.is_weighted = first_cube_has_w_counts
        _cubes_prop_.return_value = (cube_,)
        cube_set = CubeSet(None, None, None, None)

        has_weighted_counts = cube_set.has_weighted_counts

        assert has_weighted_counts == expected_value

    @pytest.mark.parametrize(("ncubes", "expected_value"), ((2, True), (1, False)))
    def it_knows_when_it_is_ca_as_0th(
        self, ncubes, expected_value, _cubes_prop_, cube_
    ):
        cubes_ = (cube_,) * ncubes
        cubes_[0].dimension_types = (DT.CA_SUBVAR,) * ncubes
        _cubes_prop_.return_value = cubes_
        cube_set = CubeSet(cubes_, None, None, None)

        is_ca_as_0th = cube_set.is_ca_as_0th

        assert is_ca_as_0th == expected_value

    @pytest.mark.parametrize(
        ("first_cube_missing_count", "expected_value"), ((34, 34), (0, 0))
    )
    def it_knows_its_missing_count(
        self, first_cube_missing_count, expected_value, _cubes_prop_, cube_
    ):
        cube_.missing = first_cube_missing_count
        _cubes_prop_.return_value = (cube_,)
        cube_set = CubeSet(None, None, None, None)

        missing_count = cube_set.missing_count

        assert missing_count == expected_value

    def it_knows_its_name(self, _cubes_prop_, cube_):
        cube_.name = "Beverage"
        _cubes_prop_.return_value = (cube_,)
        cube_set = CubeSet(None, None, None, None)

        name = cube_set.name

        assert name == "Beverage"

    @pytest.mark.parametrize(
        ("cube_partitions", "expected_value"),
        (
            # --- 3D, 2D, 1D, Nub ---
            ((_Strand, _Slice, _Slice), ((_Strand,), (_Slice,), (_Slice,))),
            ((_Slice, _Slice), ((_Slice,), (_Slice,))),
            ((_Slice,), ((_Slice,),)),
            ((_Nub,), ((_Nub,),)),
        ),
    )
    def it_provides_access_to_the_partition_sets(
        self, cube_partitions, expected_value, _cubes_prop_, cube_
    ):
        cube_.partitions = cube_partitions
        _cubes_prop_.return_value = (cube_,)
        cube_set = CubeSet(None, None, None, None)

        partition_sets = cube_set.partition_sets

        assert partition_sets == expected_value

    @pytest.mark.parametrize(
        ("population_fraction", "expected_value"),
        ((1.0, 1.0), (0.54, 0.54), (np.nan, np.nan)),
    )
    def it_has_proper_population_fraction(
        self, population_fraction, expected_value, cube_, _cubes_prop_
    ):
        cube_.population_fraction = population_fraction
        _cubes_prop_.return_value = (cube_,)
        cube_set = CubeSet(None, None, None, None)

        cubeset_population_fraction = cube_set.population_fraction

        np.testing.assert_almost_equal(cubeset_population_fraction, expected_value)

    def it_provides_access_to_its_Cube_objects_to_help(self, _iter_cubes_, cube_):
        _iter_cubes_.return_value = (c for c in (cube_, cube_, cube_))
        cube_set = CubeSet(None, None, None, None)

        cubes = cube_set._cubes

        assert cubes == (cube_, cube_, cube_)

    @pytest.mark.parametrize(
        ("is_multi_cube", "cube_0_ndim", "expected_value"),
        ((False, 1, False), (False, 0, False), (True, 1, False), (True, 0, True)),
    )
    def it_knows_whether_it_is_numeric_mean_to_help(
        self,
        _is_multi_cube_prop_,
        is_multi_cube,
        Cube_,
        cube_,
        cube_0_ndim,
        expected_value,
    ):
        _is_multi_cube_prop_.return_value = is_multi_cube
        cube_.ndim = cube_0_ndim
        Cube_.return_value = cube_
        cube_set = CubeSet(({"cube": 0}, {"cube": 1}), None, None, None)

        is_numeric_mean = cube_set._is_numeric_mean

        assert Cube_.call_args_list == ([call({"cube": 0})] if is_multi_cube else [])
        assert is_numeric_mean == expected_value

    def it_constructs_its_sequence_of_cube_objects_to_help(
        self, request, Cube_, _is_numeric_mean_prop_
    ):
        cubes_ = tuple(instance_mock(request, Cube) for _ in range(4))
        Cube_.side_effect = iter(cubes_)
        _is_numeric_mean_prop_.return_value = False
        cube_set = CubeSet(
            cube_responses=[{"cube": "resp-1"}, {"cube": "resp-2"}, {"cube": "resp-3"}],
            transforms=[{"xfrms": 1}, {"xfrms": 2}, {"xfrms": 3}],
            population=1000,
            min_base=10,
        )

        cubes = tuple(cube_set._iter_cubes())

        assert Cube_.call_args_list == [
            call(
                {"cube": "resp-1"},
                cube_idx=0,
                transforms={"xfrms": 1},
                population=1000,
                mask_size=10,
            ),
            call(
                {"cube": "resp-2"},
                cube_idx=1,
                transforms={"xfrms": 2},
                population=1000,
                mask_size=10,
            ),
            call(
                {"cube": "resp-3"},
                cube_idx=2,
                transforms={"xfrms": 3},
                population=1000,
                mask_size=10,
            ),
        ]
        assert cubes == cubes_[:3]

    def but_it_inflates_the_cubes_in_special_case_of_numeric_mean_payload(
        self, request, Cube_, cube_, _is_numeric_mean_prop_
    ):
        cubes_ = tuple(instance_mock(request, Cube) for _ in range(4))
        cube_.inflate.side_effect = iter(cubes_)
        Cube_.return_value = cube_
        _is_numeric_mean_prop_.return_value = True
        cube_set = CubeSet(
            cube_responses=[{"cube": "resp-1"}, {"cube": "resp-2"}, {"cube": "resp-3"}],
            transforms=[{"xfrms": 1}, {"xfrms": 2}, {"xfrms": 3}],
            population=1000,
            min_base=10,
        )

        cubes = tuple(cube_set._iter_cubes())

        assert Cube_.call_args_list == [
            call(
                {"cube": "resp-1"},
                cube_idx=0,
                transforms={"xfrms": 1},
                population=1000,
                mask_size=10,
            ),
            call(
                {"cube": "resp-2"},
                cube_idx=1,
                transforms={"xfrms": 2},
                population=1000,
                mask_size=10,
            ),
            call(
                {"cube": "resp-3"},
                cube_idx=2,
                transforms={"xfrms": 3},
                population=1000,
                mask_size=10,
            ),
        ]
        assert cube_.inflate.call_args_list == [call(), call(), call()]
        assert cubes == cubes_[:3]

    # fixture components ---------------------------------------------

    @pytest.fixture
    def Cube_(self, request):
        return class_mock(request, "cr.cube.cube.Cube")

    @pytest.fixture
    def cube_(self, request):
        return instance_mock(request, Cube)

    @pytest.fixture
    def _cubes_prop_(self, request):
        return property_mock(request, CubeSet, "_cubes")

    @pytest.fixture
    def _is_multi_cube_prop_(self, request):
        return property_mock(request, CubeSet, "_is_multi_cube")

    @pytest.fixture
    def _is_numeric_mean_prop_(self, request):
        return property_mock(request, CubeSet, "_is_numeric_mean")

    @pytest.fixture
    def _iter_cubes_(self, request):
        return method_mock(request, CubeSet, "_iter_cubes")


class DescribeCube(object):
    """Unit-test suite for `cr.cube.cube.Cube` object."""

    def it_provides_the_default_repr_when_enhanced_repr_fails(
        self, dimension_types_prop_
    ):
        dimension_types_prop_.return_value = [1, 2, 3]
        cube = Cube(None)

        cube_repr = cube.__repr__()

        assert cube_repr.startswith("<cr.cube.cube.Cube object at 0x")

    def it_can_inflate_itself(self, request):
        cube = Cube(
            {"result": {"dimensions": [{"other": "dim"}]}},
            cube_idx=1,
            transforms={"trans": "forms"},
            population=1000,
            mask_size=10,
        )
        # --- mock Cube only *after* constructing cube-under-test; we only want the mock
        # --- to intercept the *second* call
        inflated_cube_ = instance_mock(request, Cube)
        Cube_ = class_mock(request, "cr.cube.cube.Cube", return_value=inflated_cube_)

        inflated_cube = cube.inflate()

        Cube_.assert_called_once_with(
            {
                "result": {
                    "dimensions": [
                        {
                            "references": {"alias": "mean", "name": "mean"},
                            "type": {
                                "class": "categorical",
                                "categories": [{"id": 1, "name": "Mean"}],
                            },
                        },
                        {"other": "dim"},
                    ]
                }
            },
            1,
            {"trans": "forms"},
            1000,
            10,
        )
        assert inflated_cube is inflated_cube_

    @pytest.mark.parametrize(
        ("cube_idx_arg", "expected_value"), ((None, 0), (0, 0), (1, 1), (42, 42))
    )
    def it_knows_its_index_within_its_cube_set(self, cube_idx_arg, expected_value):
        assert Cube(None, cube_idx_arg).cube_index == expected_value

    @pytest.mark.parametrize(
        ("dim_types", "aliases", "expected_value"),
        (
            ((), (), False),
            ((DT.MR,), ("alias",), False),
            ((DT.MR, DT.MR), ("alias", "alias"), False),
            ((DT.MR, DT.CAT, DT.CAT), ("alias1", "alias2", "alias2"), False),
            ((DT.CAT, DT.MR, DT.MR), ("alias1", "alias2", "alias3"), False),
            ((DT.CAT, DT.MR, DT.MR), ("alias1", "alias2", "alias2"), True),
        ),
    )
    def it_knows_if_it_is_mr_aug(
        self, request, dim_types, aliases, expected_value, dimension_types_prop_
    ):
        property_mock(request, Cube, "ndim", return_value=len(dim_types))
        dimension_types_prop_.return_value = dim_types
        property_mock(
            request,
            Cube,
            "dimensions",
            return_value=[
                instance_mock(request, Dimension, alias=aliases[i])
                for i, _ in enumerate(dim_types)
            ],
        )
        cube = Cube(None, None, None, None)

        is_mr_aug = cube.is_mr_aug

        assert is_mr_aug is expected_value

    @pytest.mark.parametrize(
        ("dim_types", "cube_idx", "_is_single_filter_col_cube", "expected_value"),
        (
            ((), 0, False, False),
            ((), 0, True, False),
            ((), 1, True, False),
            ((DT.CA, DT.CAT), 0, False, True),
            ((DT.CA, DT.CAT), 1, True, True),
            ((DT.CAT, DT.CAT), 0, True, False),
            ((DT.CA, DT.CAT, DT.CAT), 0, True, True),
            ((DT.CA, DT.CAT, DT.CAT), 1, False, False),
        ),
    )
    def it_knows_ca_as_0th(
        self,
        request,
        dim_types,
        cube_idx,
        _is_single_filter_col_cube,
        expected_value,
        dimension_types_prop_,
    ):
        property_mock(
            request,
            Cube,
            "_is_single_filter_col_cube",
            return_value=_is_single_filter_col_cube,
        )
        dimension_types_prop_.return_value = dim_types
        cube = Cube(
            response=None,
            cube_idx=cube_idx,
            transforms=None,
            population=None,
            mask_size=0,
        )

        _ca_as_0th = cube._ca_as_0th

        assert _ca_as_0th is expected_value

    @pytest.mark.parametrize(
        ("cube_dict", "expected_value"),
        (({"result": {}}, "Untitled"), ({"result": {"title": "Hipsters"}}, "Hipsters")),
    )
    def it_knows_its_title(self, _cube_dict_prop_, cube_dict, expected_value):
        _cube_dict_prop_.return_value = cube_dict
        assert Cube(None).title == expected_value

    @pytest.mark.parametrize(
        ("cube_dict", "expected_value"),
        (({"result": {}}, False), ({"result": {"is_single_col_cube": True}}, True)),
    )
    def it_knows_if_it_is_a_single_col_filter_cube(
        self, _cube_dict_prop_, cube_dict, expected_value
    ):
        _cube_dict_prop_.return_value = cube_dict
        assert Cube(None)._is_single_filter_col_cube == expected_value

    def it_provides_access_to_the_cube_response_dict_to_help(self):
        assert Cube({"cube": "dict"})._cube_dict == {"cube": "dict"}

    @pytest.mark.parametrize(
        ("cube_response", "expected_value"),
        ((CR.CAT_X_CAT, CR.CAT_X_CAT), ({"value": "val"}, "val")),
    )
    def and_it_accepts_a_JSON_format_cube_response(self, cube_response, expected_value):
        assert Cube(cube_response)._cube_dict == expected_value

    @pytest.mark.parametrize(
        ("cube_response", "expected_value"),
        (
            (
                None,
                "Unsupported type <NoneType> provided. Cube response must be JSON "
                "(str) or dict.",
            ),
            (
                0,
                "Unsupported type <int> provided. Cube response must be JSON (str) or "
                "dict.",
            ),
        ),
    )
    def but_it_raises_on_other_cube_response_types(self, cube_response, expected_value):
        with pytest.raises(TypeError) as e:
            Cube(cube_response)._cube_dict

        assert str(e.value) == expected_value

    # fixture components ---------------------------------------------

    @pytest.fixture
    def _cube_dict_prop_(self, request):
        return property_mock(request, Cube, "_cube_dict")

    @pytest.fixture
    def dimension_types_prop_(self, request):
        return property_mock(request, Cube, "dimension_types")


class DescribeMeasures(object):
    def it_knows_the_population_fraction(self):
        cube_dict, expected_value = (
            {
                "result": {
                    "filtered": {"weighted_n": 10},
                    "unfiltered": {"weighted_n": 9},
                }
            },
            1.1111111111111112,
        )
        measures = _Measures(cube_dict, None)

        population_fraction = measures.population_fraction

        assert population_fraction == expected_value

    def but_the_fraction_is_NaN_for_unfiltered_count_zero(self):
        cube_dict, expected_value = (
            {
                "result": {
                    "filtered": {"weighted_n": 0},
                    "unfiltered": {"weighted_n": 0},
                }
            },
            np.nan,
        )
        measures = _Measures(cube_dict, None)

        population_fraction = measures.population_fraction

        np.testing.assert_equal(population_fraction, expected_value)



# encoding: utf-8

"""Integration-test suite for column/row index measure behaviors."""

from __future__ import absolute_import, division, print_function, unicode_literals

import numpy as np

from cr.cube.cube import Cube

from ..fixtures import CR


def test_cat_x_cat_slice_column_index():
    slice_ = Cube(CR.CAT_X_CAT).partitions[0]
    expected = np.array([[100, 80], [100, 120]])
    np.testing.assert_almost_equal(slice_.column_index, expected)


def test_mr_x_cat_slice_column_index():
    slice_ = Cube(CR.SELECTED_CROSSTAB_4).partitions[0]
    expected = np.array(
        [
            [95.8651525855387, 103.859044435659],
            [102.305106635277, 97.8432045727022],
            [97.6031146323114, 102.24274029149],
            [98.1029444304978, 101.829068203842],
            [114.466510106092, 86.0656684647625],
            [99.2925720053358, 100.682933745397],
        ]
    )
    np.testing.assert_almost_equal(slice_.column_index, expected)


def test_mr_x_mr_slice_column_index():
    mr_x_mr = Cube(CR.FULL_CUBE).partitions[0]
    mr_alone = Cube(CR.MR_WGTD).partitions[0]
    expected = (
        mr_x_mr.column_proportions / mr_alone._table_proportions_as_array[:, None] * 100
    )
    np.testing.assert_almost_equal(mr_x_mr.column_index, expected)


def test_cat_x_mr_slice_column_index():
    slice_ = Cube(CR.CAT_X_MR).partitions[0]
    expected = np.array(
        [[90.9090909, 106.9518717, 95.6937799], [104.4776119, 96.5759438, 102.1209741]]
    )
    np.testing.assert_almost_equal(slice_.column_index, expected)


def test_mr_x_mr_index_tables_parity_with_whaam_and_r():
    slice_ = Cube(CR.MR_X_MR_INDEX_TABLE).partitions[0]
    # Test column direction
    expected = np.array(
        [
            [192.05298013, 97.23165321, 89.68799602],
            [99.22588537, 239.38592924, 95.99096915],
            [93.52597694, 98.08689727, 182.31556654],
        ]
    )
    np.testing.assert_almost_equal(slice_.column_index, expected)


def test_mr_x_3vl_index_tables_parity_with_nssat():
    slice_ = Cube(CR.NSSAT_MR_X_3vl).partitions[0]
    # Test column direction
    expected = np.array(
        [
            [179.793686976007, 90.4924459426829],
            [182.343952039497, 88.5838171105893],
            [84.6174957937067, 101.993855386627],
            [np.nan, np.nan],
        ]
    )
    np.testing.assert_almost_equal(slice_.column_index, expected)


def test_mr_x_mr_index_tables_parity_with_nssat():
    slice_ = Cube(CR.NSSAT_MR_X_MR).partitions[0]

    # Test column direction
    expected = np.array(
        [
            [
                114.917891097666,
                94.6007480891202,
                75.7981149285497,
                41.5084915084915,
                64.5687645687646,
                581.118881118881,
                np.nan,
                0,
                np.nan,
            ],
            [
                90.0597657183839,
                95.9426026719446,
                102.497687326549,
                84.1945288753799,
                261.93853427896,
                0,
                np.nan,
                0,
                np.nan,
            ],
            [
                99.4879510762734,
                101.567130443518,
                101.446145177951,
                106.834310398025,
                86.4170866330693,
                59.8272138228942,
                np.nan,
                119.654427645788,
                np.nan,
            ],
            [np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan],
        ]
    )
    np.testing.assert_almost_equal(slice_.column_index, expected)


def test_mr_single_cat_x_mr():
    # No pruning
    slice_ = Cube(CR.MR_SINGLE_CAT_X_MR).partitions[0]
    expected = [[100, 100, np.nan]]
    np.testing.assert_array_equal(slice_.column_index, expected)
    # TODO: Implement row index table
    # np.testing.assert_array_equal(cube_slice.index_table(axis=1), expected)

    # With pruning
    transforms = {
        "rows_dimension": {"prune": True},
        "columns_dimension": {"prune": True},
    }
    slice_ = Cube(CR.MR_SINGLE_CAT_X_MR, transforms=transforms).partitions[0]
    expected = [[100, 100]]
    np.testing.assert_array_equal(slice_.column_index, expected)


def test_mr_x_mr_single_cat():
    slice_ = Cube(CR.MR_X_MR_SINGLE_CAT).partitions[0]
    expected = [[100], [100], [np.nan]]
    np.testing.assert_array_equal(slice_.column_index, expected)

    # With pruning
    transforms = {
        "rows_dimension": {"prune": True},
        "columns_dimension": {"prune": True},
    }
    slice_ = Cube(CR.MR_X_MR_SINGLE_CAT, transforms=transforms).partitions[0]
    expected = [[100], [100]]
    np.testing.assert_array_equal(slice_.column_index, expected)



# encoding: utf-8

"""Provides the Dimension class."""

import sys

if sys.version_info >= (3, 3):
    from collections.abc import Sequence
else:
    from collections import Sequence

import numpy as np

from cr.cube.enums import DIMENSION_TYPE as DT
from cr.cube.util import lazyproperty


class _BaseDimensions(Sequence):
    """Base class for dimension collections."""

    def __getitem__(self, idx_or_slice):
        """Implements indexed access."""
        return self._dimensions[idx_or_slice]

    def __iter__(self):
        """Implements (efficient) iterability."""
        return iter(self._dimensions)

    def __len__(self):
        """Implements len(elements)."""
        return len(self._dimensions)

    @lazyproperty
    def _dimensions(self):
        """tuple of dimension objects in this collection.

        This composed tuple is the source for the dimension objects in this
        collection.
        """
        raise NotImplementedError(
            "must be implemented by each subclass"
        )  # pragma: no cover


class AllDimensions(_BaseDimensions):
    """Collection containing every dimension defined in cube response."""

    def __init__(self, dimension_dicts):
        self._dimension_dicts = dimension_dicts

    @lazyproperty
    def apparent_dimensions(self):
        """_ApparentDimensions collection of the "visible" dimensions.

        The two dimensions for a multiple-response (MR) variable are
        conflated into a single dimensions in this collection.
        """
        return _ApparentDimensions(all_dimensions=self._dimensions)

    @lazyproperty
    def shape(self):
        """Tuple of int element count for each dimension.

        This corresponds to the shape of the ndarray representing the raw
        cube response values (raw meaning including missing and prunable
        elements and any MR_CAT dimensions).
        """
        return tuple(d.shape for d in self._dimensions)

    @lazyproperty
    def _dimensions(self):
        """tuple of dimension objects in this collection.

        This composed tuple is the internal source for the dimension objects
        in this collection.
        """
        return tuple(_DimensionFactory.iter_dimensions(self._dimension_dicts))


class _ApparentDimensions(_BaseDimensions):
    """Collection containing only "user" dimensions of a cube."""

    def __init__(self, all_dimensions):
        self._all_dimensions = all_dimensions

    @lazyproperty
    def _dimensions(self):
        """tuple of dimension objects in this collection.

        This composed tuple is the source for the dimension objects in this
        collection.
        """
        return tuple(d for d in self._all_dimensions if d.dimension_type != DT.MR_CAT)


class _DimensionFactory(object):
    """Produce Dimension objects of correct type from dimension-dicts.

    "type" here is primarily the `.dimension_type` value of the dimension,
    although if `Dimension` becomes an object hierarchy, this factory would
    make dimension class choices as well.
    """

    def __init__(self, dimension_dicts):
        self._dimension_dicts = dimension_dicts

    @classmethod
    def iter_dimensions(cls, dimension_dicts):
        """Generate Dimension object for each of *dimension_dicts*."""
        return cls(dimension_dicts)._iter_dimensions()

    def _iter_dimensions(self):
        """Generate Dimension object for each dimension dict."""
        return (
            Dimension(raw_dimension.dimension_dict, raw_dimension.dimension_type)
            for raw_dimension in self._raw_dimensions
        )

    @lazyproperty
    def _raw_dimensions(self):
        """Sequence of _RawDimension objects wrapping each dimension dict."""
        return tuple(
            _RawDimension(dimension_dict, self._dimension_dicts)
            for dimension_dict in self._dimension_dicts
        )


class _RawDimension(object):
    """Thin wrapper around dimension-dict to support dimension-type discovery.

    Determining dimension-type is pretty complex and requires repeated
    partial parsing of both the dimension dict and its siblings. This class
    abstracts that access for clarity.
    """

    def __init__(self, dimension_dict, dimension_dicts):
        self._dimension_dict = dimension_dict
        self._dimension_dicts = dimension_dicts

    @lazyproperty
    def dimension_dict(self):
        """dict defining this dimension in cube response."""
        return self._dimension_dict

    @lazyproperty
    def dimension_type(self):
        """Return member of DIMENSION_TYPE appropriate to dimension_dict."""
        base_type = self._base_type
        if base_type == "categorical":
            return self._resolve_categorical()
        if base_type == "enum.variable":
            return self._resolve_array_type()
        if base_type == "enum.datetime":
            return DT.DATETIME
        if base_type == "enum.numeric":
            return DT.BINNED_NUMERIC
        if base_type == "enum.text":
            return DT.TEXT
        raise NotImplementedError("unrecognized dimension type %s" % base_type)

    @lazyproperty
    def _alias(self):
        """Return str key for variable behind *dimension_dict*."""
        return self._dimension_dict["references"]["alias"]

    @lazyproperty
    def _base_type(self):
        """Return str like 'enum.numeric' representing dimension type.

        This string is a 'type.subclass' concatenation of the str keys
        used to identify the dimension type in the cube response JSON.
        The '.subclass' suffix only appears where a subtype is present.
        """
        type_class = self._dimension_dict["type"]["class"]
        if type_class == "categorical":
            return "categorical"
        if type_class == "enum":
            subclass = self._dimension_dict["type"]["subtype"]["class"]
            return "enum.%s" % subclass
        raise NotImplementedError("unexpected dimension type class '%s'" % type_class)

    @lazyproperty
    def _has_selected_category(self):
        """True if dimension-dict includes one or more selected categories.

        A "selected" category-dict is one having `'selected': True`. This
        property is only meaningful for a categorical dimension dict.
        """
        return True in {
            category.get("selected")
            for category in self._dimension_dict["type"].get("categories", [])
        }

    @lazyproperty
    def _is_array_cat(self):
        """True if a categorical dimension_dict belongs to an array pair.

        Returns True for a CA_CAT or MR_CAT dimension. Only meaningful when
        the dimension is known to be categorical (has base-type
        'categorical').
        """
        return "subreferences" in self._dimension_dict["references"]

    @lazyproperty
    def _next_raw_dimension(self):
        """_RawDimension for next *dimension_dict* in sequence or None for last.

        Returns None if this dimension is the last in sequence for this cube.
        """
        dimension_dicts = self._dimension_dicts
        this_idx = dimension_dicts.index(self._dimension_dict)
        if this_idx > len(dimension_dicts) - 2:
            return None
        return _RawDimension(dimension_dicts[this_idx + 1], self._dimension_dicts)

    def _resolve_array_type(self):
        """Return one of the ARRAY_TYPES members of DIMENSION_TYPE.

        This method distinguishes between CA and MR dimensions. The return
        value is only meaningful if the dimension is known to be of array
        type (i.e. either CA or MR, base-type 'enum.variable').
        """
        next_raw_dimension = self._next_raw_dimension
        if next_raw_dimension is None:
            return DT.CA

        is_mr_subvar = (
            next_raw_dimension._base_type == "categorical"
            and next_raw_dimension._has_selected_category
            and next_raw_dimension._alias == self._alias
        )
        return DT.MR if is_mr_subvar else DT.CA

    def _resolve_categorical(self):
        """Return one of the categorical members of DIMENSION_TYPE.

        This method distinguishes between CAT, CA_CAT, MR_CAT, and LOGICAL
        dimension types, all of which have the base type 'categorical'. The
        return value is only meaningful if the dimension is known to be one
        of the categorical types (has base-type 'categorical').
        """
        # ---an array categorical is either CA_CAT or MR_CAT---
        if self._is_array_cat:
            return DT.MR_CAT if self._has_selected_category else DT.CA_CAT

        # ---what's left is logical or plain-old categorical---
        return DT.LOGICAL if self._has_selected_category else DT.CAT


class Dimension(object):
    """Represents one dimension of a cube response.

    Each dimension represents one of the variables in a cube response. For
    example, a query to cross-tabulate snack-food preference against region
    will have two variables (snack-food preference and region) and will produce
    a two-dimensional (2D) cube response. That cube will have two of these
    dimension objects, which are accessed using
    :attr:`.CrunchCube.dimensions`.
    """

    def __init__(self, dimension_dict, dimension_type, dimension_transforms=None):
        self._dimension_dict = dimension_dict
        self._dimension_type = dimension_type
        self._dimension_transforms_arg = dimension_transforms

    @lazyproperty
    def alias(self):
        """Return the alias for the dimension if it exists, None otherwise

        This property is needed to identify one of the mandatory condition
        for a MRxItself cube.
        """
        return self._dimension_dict["references"].get("alias", None)

    @lazyproperty
    def all_elements(self):
        """_AllElements object providing cats or subvars of this dimension.

        Elements in this sequence appear in cube-result order. Display order (including
        resolution of the explicit-reordering transforms cascade) is provided by
        a separate `.display_order` attribute on _AllElements.
        """
        return _AllElements(
            self._dimension_dict["type"], self._dimension_transforms_dict
        )

    def apply_transforms(self, dimension_transforms):
        """Return a new `Dimension` object with `dimension_transforms` applied.

        The new dimension object is the same as this one in all other respects.
        """
        return Dimension(
            self._dimension_dict, self._dimension_type, dimension_transforms
        )

    @lazyproperty
    def description(self):
        """str description of this dimension."""
        # ---First authority in cascade is analysis-specific dimension transform. None
        # ---is a legitimate value, indicating suppression of any inherited subtitle.
        if "description" in self._dimension_transforms_dict:
            description = self._dimension_transforms_dict["description"]
        # ---inherited value is base dimension description---
        else:
            description = self._dimension_dict["references"].get("description")

        # ---Normalize to "" so return value is always a str and callers don't need to
        # ---deal with None as a possible return type.
        return description if description else ""

    @lazyproperty
    def dimension_type(self):
        """Member of DIMENSION_TYPE appropriate to this cube dimension."""
        return self._dimension_type

    @lazyproperty
    def display_order(self):
        """Sequence of int element indices specifying display order of elements.

        The sequence includes only valid elements; missing elements do not appear.
        Further, each index represents the document-order position of the element in the
        sequence of valid elements; missing elements are skipped in the assignment of
        indexes. The returned sequence is exhaustive; all valid elements are
        represented.

        The sequence reflects the resolved cascade of any *explicit* ordering
        transforms, but does *not* reflect any *sort* transforms, which cannot be
        resolved by the dimension. Use the `.sort` property to access any sort transform
        that may apply.

        Example with explicit-order transform:

            (3, 0, 2, 1, 4)

        Example with no explicit-order transform:

            (0, 1, 2, 3, 4)
        """
        return self.valid_elements.display_order

    @lazyproperty
    def name(self):
        """str name of this dimension, the empty string ("") if not specified."""
        references = self._dimension_dict["references"]
        # ---First authority in cascade is analysis-specific dimension transform. None
        # ---is a legitimate value, indicating suppression of any inherited title.
        if "name" in self._dimension_transforms_dict:
            name = self._dimension_transforms_dict["name"]
        # ---next authority is base dimension name---
        elif "name" in references:
            name = references["name"]
        else:
            name = references.get("alias")

        # ---Normalize None value to "" so return value is always a str and callers
        # ---don't need to deal with multiple possible return value types.
        return name if name else ""

    @lazyproperty
    def numeric_values(self):
        """tuple of numeric values for valid elements of this dimension.

        Each category of a categorical variable can be assigned a *numeric
        value*. For example, one might assign `like=1, dislike=-1,
        neutral=0`. These numeric mappings allow quantitative operations
        (such as mean) to be applied to what now forms a *scale* (in this
        example, a scale of preference).

        The numeric values appear in the same order as the
        categories/elements of this dimension. Each element is represented by
        a value, but an element with no numeric value appears as `np.nan` in
        the returned list.
        """
        return tuple(element.numeric_value for element in self.valid_elements)

    @lazyproperty
    def prune(self):
        """True if empty elements should be automatically hidden on this dimension."""
        prune = self._dimension_transforms_dict.get("prune")
        if prune is True:
            return True
        return False

    @lazyproperty
    def shape(self):
        return len(self.all_elements)

    @lazyproperty
    def sort(self):
        """A _BaseSort-subclass object or None, describing the applied sort method.

        This value is None if no sort transform was specified for this dimension.
        Currently that is its only possible value. The returned sort object describes
        the sort method which can include sorting on the value of an opposing element or
        on the margin and specify ascending or descending order.
        """
        return None  # pragma: no cover

    @lazyproperty
    def subtotals(self):
        """_Subtotals sequence object for this dimension.

        Each item in the sequence is a _Subtotal object specifying a subtotal, including
        its addends and anchor.
        """
        # ---insertions in dimension-transforms override those on dimension itself---
        insertion_dicts = self._dimension_transforms_dict.get("insertions")
        if insertion_dicts is not None:
            return _Subtotals(insertion_dicts, self.valid_elements, self.prune)

        # ---otherwise, insertions defined as default transforms apply---
        view = self._dimension_dict.get("references", {}).get("view", {})
        # ---view can be both None and {}, thus the edge case.---
        insertion_dicts = (
            [] if view is None else view.get("transform", {}).get("insertions", [])
        )
        return _Subtotals(insertion_dicts, self.valid_elements, self.prune)

    @lazyproperty
    def valid_elements(self):
        """_Elements object providing access to non-missing elements.

        Any categories or subvariables representing missing data are excluded
        from the collection; this sequence represents a subset of that
        provided by `.all_elements`.
        """
        return self.all_elements.valid_elements

    @lazyproperty
    def _dimension_transforms_dict(self):
        """dict complying with dimension-transforms schema for this dimension.

        This value derives from the `dimension_transforms` argument passed on
        construction. When that argument is not specified, this value is an empty dict.
        """
        return (
            self._dimension_transforms_arg
            if self._dimension_transforms_arg is not None
            else {}
        )


class _BaseElements(Sequence):
    """Base class for element sequence containers."""

    def __getitem__(self, idx_or_slice):
        """Implements indexed access."""
        return self._elements[idx_or_slice]

    def __iter__(self):
        """Implements (efficient) iterability."""
        return iter(self._elements)

    def __len__(self):
        """Implements len(elements)."""
        return len(self._elements)

    @lazyproperty
    def element_ids(self):
        """tuple of element-id for each element in collection.

        Element ids appear in the order they occur in the cube response.
        """
        return tuple(element.element_id for element in self._elements)

    @lazyproperty
    def element_idxs(self):
        """tuple of element-index for each element in collection.

        Element index values represent the position of this element in the
        dimension-dict it came from. In the case of an _AllElements object,
        it will simply be a tuple(range(len(all_elements))).
        """
        return tuple(element.index for element in self._elements)

    def get_by_id(self, element_id):
        """Return _Element object identified by *element_id*.

        Raises KeyError if not found. Only elements known to this collection
        are accessible for lookup. For example, a _ValidElements object will
        raise KeyError for the id of a missing element.
        """
        return self._elements_by_id[element_id]

    @lazyproperty
    def _elements(self):
        """tuple storing actual sequence of element objects.

        Must be implemented by each subclass.
        """
        raise NotImplementedError("must be implemented by each subclass")

    @lazyproperty
    def _elements_by_id(self):
        """dict mapping each element by its id."""
        return {element.element_id: element for element in self._elements}


class _AllElements(_BaseElements):
    """Sequence of _BaseElement subclass objects for a dimension.

    Each element is either a category or a subvariable.
    """

    def __init__(self, type_dict, dimension_transforms_dict):
        self._type_dict = type_dict
        self._dimension_transforms_dict = dimension_transforms_dict

    @lazyproperty
    def valid_elements(self):
        """_ValidElements object containing only non-missing elements."""
        return _ValidElements(self._elements, self._dimension_transforms_dict)

    @lazyproperty
    def _element_dicts(self):
        """Sequence of element-dicts for this dimension, taken from cube-result."""
        return (
            self._type_dict["categories"]
            if self._type_dict["class"] == "categorical"
            else self._type_dict["elements"]
        )

    @lazyproperty
    def _elements(self):
        """tuple storing actual sequence of element objects."""
        return tuple(
            _Element(
                element_dict,
                idx,
                _ElementTransforms(element_transforms_dict, self._prune),
            )
            for (
                idx,
                element_dict,
                element_transforms_dict,
            ) in self._iter_element_makings()
        )

    def _iter_element_makings(self):
        """Generate tuple of values needed to construct each element object.

        An (idx, element_dict, element_transforms_dict) tuple is generated for each
        element in this dimension, in the order they appear in the cube-result. All
        elements are included (including missing).
        """
        element_dicts = self._element_dicts
        elements_transforms = self._dimension_transforms_dict.get("elements", {})
        for idx, element_dict in enumerate(element_dicts):
            element_id = element_dict["id"]
            # TODO: Each element transforms dict is keyed by the str() version of it int
            # value as a consequence of JSON serialization (which does not allow
            # non-string keys). Hence the str(element_id) here.
            element_transforms_dict = elements_transforms.get(
                element_id, elements_transforms.get(str(element_id), {})
            )
            yield idx, element_dict, element_transforms_dict

    @lazyproperty
    def _prune(self):
        """True if empty elements in this dimension should be automatically hidden."""
        return True if self._dimension_transforms_dict.get("prune") is True else False


class _ValidElements(_BaseElements):
    """Sequence of non-missing element objects for a dimension.

    *all_elements* is an instance of _AllElements containing all the elements
    of a dimension. This object is only intended to be constructed by
    _AllElements.valid_elements and there should be no reason to construct it
    directly.
    """

    def __init__(self, all_elements, dimension_transforms_dict):
        self._all_elements = all_elements
        self._dimension_transforms_dict = dimension_transforms_dict

    @lazyproperty
    def display_order(self):
        """Sequence of int element-idx reflecting order in which to display elements.

        This order reflects the application of any explicit element-order transforms,
        including resolution of any cascade. It does *not* reflect the results of
        a *sort* transform, which can only be resolved at a higher level, where vector
        values are known.
        """
        return (
            self._explicit_order
            if self._explicit_order
            else tuple(range(len(self._elements)))
        )

    @lazyproperty
    def _elements(self):
        """tuple containing actual sequence of element objects."""
        return tuple(element for element in self._all_elements if not element.missing)

    @lazyproperty
    def _explicit_order(self):
        """Sequence of int element-idx or None, reflecting explicit-order transform.

        This value is None if no explicit-order transform is specified. Otherwise, it is
        an exhaustive collection of (valid) element offsets, in the order specified (and
        in some cases implied) by the order transform.
        """
        # ---get order transform if any, aborting if no explicit order transform---
        order_dict = self._dimension_transforms_dict.get("order", {})
        order_type = order_dict.get("type")
        ordered_element_ids = order_dict.get("element_ids")
        if order_type != "explicit" or not isinstance(ordered_element_ids, list):
            return None

        # ---list like [0, 1, 2, -1], perhaps ["0001", "0002", etc.], reflecting element
        # ---ids in the order they appear in the cube result. We'll use this to map
        # ---element-id to its index in the valid-elements sequence.
        cube_result_order = tuple(element.element_id for element in self)
        # ---this is a copy of the same, but we're going to mutate this one. This is
        # ---required to implement the "no-duplicates" behavior.
        remaining_element_ids = list(cube_result_order)

        # ---we'll collect the results in this---
        ordered_idxs = []
        # ---append idx of each element mentioned by id in transform, in order. Remove
        # ---each element-id from remaining as we go to keep track of dups and leftovers
        for element_id in ordered_element_ids:
            # ---An element-id appearing in transform but not in dimension is ignored.
            # ---Also, a duplicated element-id is only used on first encounter.
            if element_id not in remaining_element_ids:
                continue
            ordered_idxs.append(cube_result_order.index(element_id))
            remaining_element_ids.remove(element_id)

        # ---any remaining elements are tacked onto the end of the list in the order
        # ---they originally appeared in the cube-result.
        for element_id in remaining_element_ids:
            ordered_idxs.append(cube_result_order.index(element_id))

        return tuple(ordered_idxs)


class _Element(object):
    """A category or subvariable of a dimension.

    This object resolves the transform cascade for element-level transforms.
    """

    def __init__(self, element_dict, index, element_transforms):
        self._element_dict = element_dict
        self._index = index
        self._element_transforms = element_transforms

    @lazyproperty
    def element_id(self):
        """int identifier for this category or subvariable."""
        return self._element_dict["id"]

    @lazyproperty
    def fill(self):
        """str RGB color like "#af032d" or None if not specified.

        A value of None indicates the default fill should be used for this element.
        A str value must be a hash character ("#") followed by six hexadecimal digits.
        Three-character color contractions (like "#D07") are not valid.
        """
        # ---first authority is fill transform in element transforms. The base value is
        # ---the only prior authority and that is None (use default fill color).
        return self._element_transforms.fill

    @lazyproperty
    def index(self):
        """int offset at which this element appears in dimension.

        This position is based upon the document position of this element in
        the cube response. No adjustment for missing elements is made.
        """
        return self._index

    @lazyproperty
    def is_hidden(self):
        """True if this element is explicitly hidden in this analysis."""
        # ---first authority is hide transform in element transforms---
        # ---default is not hidden, there is currently no prior-level hide transform---
        return {True: True, False: False, None: False}[self._element_transforms.hide]

    @lazyproperty
    def label(self):
        """str display-name for this element.

        This value is the empty string when no value has been specified or display of
        the name has been suppressed.

        This property handles elements for variables of all types, including
        categorical, array (subvariable), numeric, datetime and text.
        """
        # ---first authority is name transform in element transforms---
        name = self._element_transforms.name
        if name is not None:
            return name if name else ""

        # ---otherwise base-name from element-dict is used---
        element_dict = self._element_dict

        # ---category elements have a name item---
        if "name" in element_dict:
            name = element_dict["name"]
            return name if name else ""

        # ---other types are more complicated---
        value = element_dict.get("value")
        type_name = type(value).__name__

        if type_name == "NoneType":
            return ""

        if type_name == "list":
            # ---like '10-15' or 'A-F'---
            return "-".join([str(item) for item in value])

        if type_name in ("float", "int"):
            return str(value)

        if type_name in ("str", "unicode"):
            return value

        # ---For CA and MR subvar dimensions---
        name = value.get("references", {}).get("name")
        return name if name else ""

    @lazyproperty
    def missing(self):
        """True if this element represents missing data.

        False if this category or subvariable represents valid (collected)
        data.
        """
        return bool(self._element_dict.get("missing"))

    @lazyproperty
    def numeric_value(self):
        """Numeric value assigned to element by user, np.nan if absent."""
        numeric_value = self._element_dict.get("numeric_value")
        return np.nan if numeric_value is None else numeric_value

    @lazyproperty
    def prune(self):
        """True if this element should be hidden when empty, False otherwise."""
        return self._element_transforms.prune


class _ElementTransforms(object):
    """A value object providing convenient access to transforms for a single element."""

    def __init__(self, element_transforms_dict, prune):
        self._element_transforms_dict = element_transforms_dict
        self._prune = prune

    @lazyproperty
    def fill(self):
        """str RGB color like "#af032d" or None if not specified.

        A value of None indicates no fill transform was specified for this element.
        A str value must be a hash character ("#") followed by six hexadecimal digits.
        Three-character color contractions (like "#D07") are not valid.
        """

        fill = self._element_transforms_dict.get("fill")

        if not fill:
            return None
        return fill

    @lazyproperty
    def hide(self):
        """Tri-value, True if this element has been explicitly hidden in this analysis.

        False overrides any prior "hide" transform with "show" and None signifies
        "inherit".
        """
        hide = self._element_transforms_dict.get("hide")
        # ---cover all of "omitted", "==None", and odd "==[]" or "==''" cases---
        if hide is True:
            return True
        if hide is False:
            return False
        return None

    @lazyproperty
    def name(self):
        """str display-name for this element or None if not specified."""
        # ---if "name": element is omitted, no transform is specified---
        if "name" not in self._element_transforms_dict:
            return None
        # ---otherwise normalize value to str, with an explicit value of None, [], 0,
        # ---etc. becoming the empty string ("").
        name = self._element_transforms_dict["name"]
        return str(name) if name else ""

    @lazyproperty
    def prune(self):
        """True if this element should be hidden when empty."""
        return self._prune


class _Subtotals(Sequence):
    """Sequence of _Subtotal objects for a dimension.

    Each _Subtotal object represents a "subtotal" insertion transformation
    defined for the dimension.

    A subtotal can only involve valid (i.e. non-missing) elements.
    """

    def __init__(self, insertion_dicts, valid_elements, prune):
        self._insertion_dicts = insertion_dicts
        self._valid_elements = valid_elements
        self._prune = prune

    def __getitem__(self, idx_or_slice):
        """Implements indexed access."""
        return self._subtotals[idx_or_slice]

    def __iter__(self):
        """Implements (efficient) iterability."""
        return iter(self._subtotals)

    def __len__(self):
        """Implements len(subtotals)."""
        return len(self._subtotals)

    @lazyproperty
    def anchor_idxs(self):
        """List of int indicating the actual position of the subtotals."""
        # TODO: this appears wrong, using an anchor (element_id) as an index (offset of
        # anchor element within elements). This may appear to work if category-ids are
        # assigned consecutively starting at 1, but will not work in the general case.
        # I'll bet it also doesn't work when more than one subtotal is anchored to the
        # same element-id.

        def occurrences(s, lst):
            return (i for i, e in enumerate(lst) if e == s)

        anchors = self._anchors
        top_anchors = list(occurrences("top", anchors))
        int_anchors = [a + len(top_anchors) for a in anchors if isinstance(a, int)]
        bottom_anchors = [
            a + len(self._valid_elements) + len(top_anchors + int_anchors)
            for a, _ in enumerate(list(occurrences("bottom", anchors)))
        ]
        final_anchors = top_anchors + int_anchors + bottom_anchors
        return final_anchors

    def iter_for_anchor(self, anchor):
        """Generate each subtotal having matching *anchor*."""
        return (subtotal for subtotal in self._subtotals if subtotal.anchor == anchor)

    @lazyproperty
    def _anchors(self):
        """List of int or str indicating element under which to insert this subtotal."""
        return list(
            _Subtotal(subtotal_dict, self._valid_elements, self._prune).anchor
            for subtotal_dict in self._iter_valid_subtotal_dicts()
        )

    @lazyproperty
    def _element_ids(self):
        """frozenset of int id of each non-missing cat or subvar in dim."""
        return frozenset(self._valid_elements.element_ids)

    def _iter_valid_subtotal_dicts(self):
        """Generate each insertion dict that represents a valid subtotal."""
        for insertion_dict in self._insertion_dicts:
            # ---skip any non-dicts---
            if not isinstance(insertion_dict, dict):
                continue

            # ---skip any non-subtotal insertions---
            if insertion_dict.get("function") != "subtotal":
                continue

            # ---skip any hidden insertions---
            if insertion_dict.get("hide") is True:
                continue

            # ---skip any malformed subtotal-dicts---
            if not {"anchor", "args", "name"}.issubset(insertion_dict.keys()):
                continue

            # ---skip if doesn't reference at least one non-missing element---
            if not self._element_ids.intersection(insertion_dict["args"]):
                continue

            # ---an insertion-dict that successfully runs this gauntlet
            # ---is a valid subtotal dict
            yield insertion_dict

    @lazyproperty
    def _subtotals(self):
        """Composed tuple storing actual sequence of _Subtotal objects."""
        return tuple(
            _Subtotal(subtotal_dict, self._valid_elements, self._prune)
            for subtotal_dict in self._iter_valid_subtotal_dicts()
        )


class _Subtotal(object):
    """A subtotal insertion on a cube dimension."""

    def __init__(self, subtotal_dict, valid_elements, prune):
        self._subtotal_dict = subtotal_dict
        self._valid_elements = valid_elements
        self._prune = prune

    @lazyproperty
    def anchor(self):
        """int or str indicating element under which to insert this subtotal.

        An int anchor is the id of the dimension element (category or
        subvariable) under which to place this subtotal. The return value can
        also be one of 'top' or 'bottom'.

        The return value defaults to 'bottom' for an anchor referring to an
        element that is no longer present in the dimension or an element that
        represents missing data.
        """
        anchor = self._subtotal_dict["anchor"]

        if anchor is None:
            # In the case of undefined anchor default to "bottom"
            return "bottom"

        try:
            anchor = int(anchor)
            if anchor not in self._valid_elements.element_ids:
                # In the case of a non-valid int id, default to "bottom"
                return "bottom"
            return anchor
        except (TypeError, ValueError):
            return anchor.lower()

    @lazyproperty
    def addend_ids(self):
        """tuple of int ids of elements contributing to this subtotal.

        Any element id not present in the dimension or present but
        representing missing data is excluded.
        """
        return tuple(
            arg
            for arg in self._subtotal_dict.get("args", [])
            if arg in self._valid_elements.element_ids
        )

    @lazyproperty
    def addend_idxs(self):
        """ndarray of int base-element offsets contributing to this subtotal.

        Suitable for directly indexing a numpy array object (such as base values or
        margin) to extract the addend values for this subtotal.
        """
        addend_ids = self.addend_ids
        return np.fromiter(
            (
                idx
                for idx, vector in enumerate(self._valid_elements)
                if vector.element_id in addend_ids
            ),
            dtype=int,
        )

    @lazyproperty
    def label(self):
        """str display name for this subtotal, suitable for use as label."""
        name = self._subtotal_dict.get("name")
        return name if name else ""

    @lazyproperty
    def prune(self):
        """True if this subtotal should not appear when empty."""
        return self._prune



# encoding: utf-8

"""Unit test suite for `cr.cube.noa.smoothing` module."""

import numpy as np
import pytest

from cr.cube.cubepart import _Slice
from cr.cube.noa.smoothing import SingleSidedMovingAvgSmoother

from ..unitutil import initializer_mock, instance_mock, property_mock


class DescribeSingleSideMovingAvgSmoother(object):
    def it_constructs_single_sided_moving_avg_to_help(self, request, slice_):
        _init_ = initializer_mock(request, SingleSidedMovingAvgSmoother)

        single_sided_miving_avg = SingleSidedMovingAvgSmoother(slice_, {})

        _init_.assert_called_once_with(single_sided_miving_avg, slice_, {})
        assert isinstance(single_sided_miving_avg, SingleSidedMovingAvgSmoother)

    @pytest.mark.parametrize(
        "categories, expected_value",
        (
            ([], False),
            ([{"date": "2019-10-10"}, {}], False),
            ([{"date": "2019-10-10"}, {"missing": False}], False),
            ([{"date": "2019-10-10"}, {"date": "2019-11-10"}], True),
            ([{"date": "2019-10-10"}, {"date": "2019-11-10"}, {"missing": True}], True),
        ),
    )
    def it_knows_its_is_cat_date_prop(self, categories, expected_value, slice_):
        dimension_dict = {"type": {"categories": categories}}
        slice_.smoothed_dimension_dict = dimension_dict

        _is_cat_date = SingleSidedMovingAvgSmoother(slice_, {})._is_cat_date

        assert _is_cat_date is expected_value

    @pytest.mark.parametrize(
        "measure_expr, expected_value",
        (
            ({"base_measure": "col_percent"}, "column_percentages"),
            ({"base_measure": "col_index"}, "column_index"),
            ({"base_measure": "scale_mean"}, "scale_means_row"),
            ({"base_measure": "mean"}, "means"),
        ),
    )
    def it_knows_its_base_measure_mapping(self, measure_expr, expected_value):
        base_measure = SingleSidedMovingAvgSmoother(None, measure_expr)._base_measure

        assert base_measure == expected_value

    @pytest.mark.parametrize(
        "measure_expr, expected_value", (({}, None), ({"base_measure": "foo"}, "foo"))
    )
    def but_it_raises_an_exception_when_it_is_not_valid(
        self, measure_expr, expected_value
    ):
        with pytest.raises(ValueError) as err:
            SingleSidedMovingAvgSmoother(None, measure_expr)._base_measure

        assert str(err.value) == (
            "Base measure not recognized. Allowed values: 'col_percent', "
            "'col_index', 'mean', 'scale_mean', got: {}.".format(expected_value)
        )

    @pytest.mark.parametrize("window, is_cat_date", ((2, True), (3, True)))
    def it_knows_when_it_can_apply_smoothing(
        self, _base_values_prop_, _window_prop_, window, _is_cat_date_prop_, is_cat_date
    ):
        _base_values_prop_.return_value = np.random.rand(3, 5)
        _is_cat_date_prop_.return_value = is_cat_date
        _window_prop_.return_value = window

        assert SingleSidedMovingAvgSmoother(None, {})._can_smooth is True

    @pytest.mark.parametrize("shape, window", (((2, 3), 0), ((2, 3), 1), ((2, 3), 4)))
    def but_it_warns_when_window_is_invalid(
        self, _base_values_prop_, shape, _window_prop_, window, _is_cat_date_prop_
    ):
        _base_values_prop_.return_value = np.random.rand(*shape)
        _window_prop_.return_value = window
        _is_cat_date_prop_.return_value = True
        expected_warning_regex = (
            r"No smoothing performed. Smoothing window must be between 2 and the "
            r"number of periods \(%d\), got %d" % (shape[-1], window)
        )

        with pytest.warns(UserWarning, match=expected_warning_regex):
            can_smooth = SingleSidedMovingAvgSmoother(None, {})._can_smooth

        assert can_smooth is False

    @pytest.mark.parametrize("shape, window", (((2, 3), 2), ((2, 3), 3), ((2, 5), 4)))
    def and_it_warns_when_dim_is_not_a_categorical_date(
        self, _base_values_prop_, shape, _window_prop_, window, _is_cat_date_prop_
    ):
        _base_values_prop_.return_value = np.random.rand(*shape)
        _window_prop_.return_value = window
        _is_cat_date_prop_.return_value = False
        expected_warning_regex = (
            r"No smoothing performed. Column dimension must be a categorical date."
        )

        with pytest.warns(UserWarning, match=expected_warning_regex):
            can_smooth = SingleSidedMovingAvgSmoother(None, {})._can_smooth

        assert can_smooth is False

    @pytest.mark.parametrize(
        "base_values, window, expected_value",
        (
            (np.array([1, 4, 6, 7, 8, 10]), 3, [3.666667, 5.666667, 7.0, 8.333333]),
            (np.array([[3, 4, 5, 6], [7, 8, 9, 1]]), 1, [[3, 4, 5, 6], [7, 8, 9, 1]]),
            (
                np.array([[3, 4, 5, 6], [7, 8, 9, 1]]),
                2,
                [[3.5, 4.5, 5.5], [7.5, 8.5, 5.0]],
            ),
            (np.array([[3, 4, 5, 6], [7, 8, 9, 1]]), 3, [[4.0, 5.0], [8.0, 6.0]]),
            (np.array([[3, 4, 5, 6], [7, 8, 9, 1]]), 4, [[4.5], [6.25]]),
        ),
    )
    def it_knows_its_smoothed_values(
        self, _base_values_prop_, base_values, window, expected_value
    ):
        _base_values_prop_.return_value = base_values
        function_spec = {"window": window}

        _smoothed_values_ = SingleSidedMovingAvgSmoother(
            None, function_spec
        )._smoothed_values

        np.testing.assert_array_almost_equal(_smoothed_values_, expected_value)

    def it_knows_its_base_values(self, slice_):
        slice_.column_percentages = np.array([[0.3, 0.4], [0.5, 0.6]])
        function_spec = {"base_measure": "col_percent"}

        _base_vaules = SingleSidedMovingAvgSmoother(slice_, function_spec)._base_values

        np.testing.assert_array_almost_equal(_base_vaules, [[0.3, 0.4], [0.5, 0.6]])

    @pytest.mark.parametrize(
        "base_values, show_smoothing, smoothed_values, expected_value",
        (
            (
                np.array([1, 4, 6, 7, 8, 10]),
                True,
                np.array([3.666667, 5.666667, 7.0, 8.333333]),
                [np.nan, np.nan, 3.666667, 5.666667, 7.0, 8.333333],
            ),
            (np.array([1, 4, 6, 7, 8, 10]), False, [], [1, 4, 6, 7, 8, 10]),
        ),
    )
    def it_provides_complete_smoothed_values(
        self,
        _smoothed_values_prop_,
        _base_values_prop_,
        _can_smooth_prop_,
        base_values,
        show_smoothing,
        smoothed_values,
        expected_value,
    ):
        _base_values_prop_.return_value = base_values
        _smoothed_values_prop_.return_value = smoothed_values
        _can_smooth_prop_.return_value = show_smoothing

        values = SingleSidedMovingAvgSmoother(None, {}).values

        np.testing.assert_array_almost_equal(values, expected_value)

    @pytest.mark.parametrize(
        "measure_expr, expected_value", (({"window": 3}, 3), ({"window": 2}, 2))
    )
    def it_knwos_its_window_parameter(self, measure_expr, expected_value):
        window = SingleSidedMovingAvgSmoother(None, measure_expr)._window

        assert window == expected_value

    def but_it_raises_an_exception_when_it_is_None(self):
        with pytest.raises(ValueError) as err:
            SingleSidedMovingAvgSmoother(None, {})._window

        assert str(err.value) == "Window parameter cannot be None."

    # fixture components ---------------------------------------------

    @pytest.fixture
    def slice_(self, request):
        return instance_mock(request, _Slice)

    @pytest.fixture
    def _is_cat_date_prop_(self, request):
        return property_mock(request, SingleSidedMovingAvgSmoother, "_is_cat_date")

    @pytest.fixture
    def _base_values_prop_(self, request):
        return property_mock(request, SingleSidedMovingAvgSmoother, "_base_values")

    @pytest.fixture
    def _smoothed_values_prop_(self, request):
        return property_mock(request, SingleSidedMovingAvgSmoother, "_smoothed_values")

    @pytest.fixture
    def _can_smooth_prop_(self, request):
        return property_mock(request, SingleSidedMovingAvgSmoother, "_can_smooth")

    @pytest.fixture
    def _window_prop_(self, request):
        return property_mock(request, SingleSidedMovingAvgSmoother, "_window")



# encoding: utf-8

"""Cube-partition objects.

A cube-partition allows cubes of various dimensionality to be processed in a uniform
way. For example, a 2D cube becomes a `_Slice` object, but a 3D cube is "sliced" into
a sequence of `_Slice` objects; a `_Slice` object corresponds to a crosstab, and can be
operated on consistently whether it is "alone" or one of a sequence that came from a 3D
cube.

Cube-partition objects are typically used for display of secondary analysis, often in an
Excel spreadsheet but also other formats.

The three types of cube partition are the *slice*, *strand*, and *nub*, which are 2D,
1D, and 0D respectively.
"""

from __future__ import division

import numpy as np

from cr.cube.enums import DIMENSION_TYPE as DT
from cr.cube.min_base_size_mask import MinBaseSizeMask
from cr.cube.measures.pairwise_significance import PairwiseSignificance
from cr.cube.matrix import TransformedMatrix
from cr.cube.noa.smoothing import SingleSidedMovingAvgSmoother
from cr.cube.scalar import MeansScalar
from cr.cube.stripe import TransformedStripe
from cr.cube.util import lazyproperty

# ---This is the quantile of the normal Cumulative Distribution Function (CDF) at
# ---probability 97.5% (p=.975), since the computed confidence interval
# ---is ±2.5% (.025) on each side of the CDF.
Z_975 = 1.959964


class CubePartition(object):
    """A slice, a strand, or a nub drawn from a cube-response.

    These represent 2, 1, or 0 dimensions of a cube, respectively.
    """

    def __init__(self, cube, transforms=None):
        self._cube = cube
        self._transforms_arg = transforms

    @classmethod
    def factory(
        cls,
        cube,
        slice_idx=0,
        transforms=None,
        population=None,
        ca_as_0th=None,
        mask_size=0,
    ):
        """Return slice, strand, or nub object appropriate to passed parameters."""
        if cube.ndim == 0:
            return _Nub(cube)
        if cube.ndim == 1 or ca_as_0th:
            return _Strand(
                cube, transforms, population, ca_as_0th, slice_idx, mask_size
            )
        return _Slice(cube, slice_idx, transforms, population, mask_size)

    @lazyproperty
    def cube_index(self):
        """Offset of this partition's cube in its CubeSet.

        Used to differentiate certain partitions like a filtered rows-summary strand.
        """
        return self._cube.cube_index

    @lazyproperty
    def cube_is_mr_aug(self):
        return False

    @lazyproperty
    def cube_row_dimension_type(self):
        """Member of `cr.cube.enum.DIMENSION_TYPE` for cube row dimension"""
        return self._cube.dimensions[0].dimension_type

    @lazyproperty
    def dimension_types(self):
        """Sequence of member of `cr.cube.enum.DIMENSION_TYPE` for each dimension.

        Items appear in rows-dimension, columns-dimension order.
        """
        return tuple(d.dimension_type for d in self._dimensions)

    def evaluate(self, measure_expr):
        """Return 1D/2D ndarray result of evaluating `measure_expr`.

        `measure_expr` contains the function to apply and its parameters, like::

            {
                 "function": "one_sided_moving_avg",
                 "base_measure": "col_percent",
                 "window": 3
            }

        This expression specifies application of the `one_sided_moving_avg` function to
        the `col_percent` base-measure, with a sliding window of 3 periods.
        """
        function = measure_expr.get("function", None)
        if function != "one_sided_moving_avg":
            raise NotImplementedError("Function {} is not available.".format(function))
        return SingleSidedMovingAvgSmoother(self, measure_expr).values

    @lazyproperty
    def has_means(self):
        """True if cube-result includes means values."""
        return self._cube.has_means

    @lazyproperty
    def ndim(self):
        """int count of dimensions for this partition."""
        return len(self._dimensions)

    @lazyproperty
    def population_fraction(self):
        """population fraction of the cube"""
        return self._cube.population_fraction

    @lazyproperty
    def shape(self):
        """Tuple of int vector counts for this partition.

        Not to be confused with `numpy.ndarray.shape`, this represent the count of rows
        and columns respectively, in this partition. It does not necessarily represent
        the shape of any underlying `numpy.ndarray` object that may arise in the
        implementation of the cube partition. In particular, the value of any count in
        the shape can be zero.

        A _Slice has a shape like (2, 3) representing (row-count, col-count). A _Strand
        has a shape like (5,) which represents its row-count. The shape of a _Nub is
        unconditionally () (an empty tuple).
        """
        raise NotImplementedError(
            "must be implemented by each subclass"
        )  # pragma: no cover

    @lazyproperty
    def variable_name(self):
        """str representing the name of the superheading variable."""
        return self._dimensions[0 if self.ndim < 2 else 1].name

    @lazyproperty
    def _alpha(self):
        """float confidence-interval threshold for pairwise-t (sig) tests."""
        return self._alpha_values[0]

    @lazyproperty
    def _alpha_alt(self):
        """Alternate float confidence-interval threshold or None.

        This is an optional secondary confidence interval allowing two-level
        significance testing. Value is None if no alternate alpha was specified by user.
        """
        return self._alpha_values[1]

    @lazyproperty
    def _alpha_values(self):
        """Pair (tuple) of confidence-interval thresholds to be used for t-tests.

        The second value is optional and is None when no secondary alpha value was
        defined for the cube-set.
        """
        value = self._transforms_dict.get("pairwise_indices", {}).get("alpha")

        # --- handle omitted, None, [], (), {}, "", 0, and 0.0 cases ---
        if not value:
            return (0.05, None)

        # --- reject invalid types ---
        if not isinstance(value, (float, list, tuple)):
            raise TypeError(
                "transforms.pairwise_indices.alpha, when defined, must be a list of 1 "
                "or 2 float values between 0.0 and 1.0 exclusive. Got %r" % value
            )

        # --- legacy float "by-itself" case ---
        if isinstance(value, float):
            if not 0.0 < value < 1.0:
                raise ValueError(
                    "alpha value, when provided, must be between 0.0 and 1.0 "
                    "exclusive. Got %r" % value
                )
            return (value, None)

        # --- sequence case ---
        for x in value[:2]:
            if not isinstance(x, float) or not 0.0 < x < 1.0:
                raise ValueError(
                    "transforms.pairwise_indices.alpha must be a list of 1 or 2 float "
                    "values between 0.0 and 1.0 exclusive. Got %r" % value
                )

        if len(value) == 1:
            return (value[0], None)

        return tuple(sorted(value[:2]))

    @lazyproperty
    def _dimensions(self):
        """tuple of Dimension object for each dimension in cube-partition."""
        raise NotImplementedError(
            "must be implemented by each subclass"
        )  # pragma: no cover

    @lazyproperty
    def _only_larger(self):
        """True if only the larger of reciprocal pairwise-t values should appear.

        In general, pairwise-t tests are reciprocal. That is, if A is significant with
        respect to B, then B is significant with respect to A. Having a letter in both
        columns can produce a cluttered appearance. When this flag is set by the user,
        only the cell in the reciprocal pair having the largest value gets a letter.
        Defaults to True unless explicitly set False.
        """
        return (
            False
            if self._transforms_dict.get("pairwise_indices", {}).get(
                "only_larger", True
            )
            is False
            else True
        )

    @lazyproperty
    def _transforms_dict(self):
        """dict holding transforms for this partition, provided as `transforms` arg.

        This value is an empty dict (`{}`) when no transforms were specified on
        construction.
        """
        return {} if self._transforms_arg is None else self._transforms_arg


class _Slice(CubePartition):
    """2D cube partition.

    A slice represents the cross-tabulation of two dimensions, often, but not
    necessarily contributed by two different variables. A single CA variable has two
    dimensions which can be crosstabbed in a slice.
    """

    def __init__(self, cube, slice_idx, transforms, population, mask_size):
        super(_Slice, self).__init__(cube, transforms)
        self._slice_idx = slice_idx
        self._population = population
        self._mask_size = mask_size

    # ---interface ---------------------------------------------------

    @lazyproperty
    def column_base(self):
        return np.array([column.base for column in self._matrix.columns]).T

    @lazyproperty
    def column_index(self):
        """ndarray of column index percentages.

        The index values represent the difference of the percentages to the
        corresponding baseline values. The baseline values are the univariate
        percentages of the corresponding variable.
        """
        return np.array([row.column_index for row in self._matrix.rows])

    @lazyproperty
    def column_labels(self):
        """Sequence of str column element names suitable for use as column headings."""
        return tuple(column.label for column in self._matrix.columns)

    @lazyproperty
    def column_percentages(self):
        return self.column_proportions * 100

    @lazyproperty
    def column_proportions(self):
        return np.array([col.proportions for col in self._matrix.columns]).T

    @lazyproperty
    def columns_dimension_name(self):
        """str name assigned to columns-dimension.

        Reflects the resolved dimension-name transform cascade.
        """
        return self._columns_dimension.name

    @lazyproperty
    def columns_dimension_type(self):
        """Member of `cr.cube.enum.DIMENSION_TYPE` describing columns dimension."""
        return self._columns_dimension.dimension_type

    @lazyproperty
    def columns_margin(self):
        return np.array([column.margin for column in self._matrix.columns]).T

    @lazyproperty
    def columns_percentages_moe(self):
        """1D/2D np.float64 ndarray of margin-of-error (MoE) for columns percentages.

        The values are represented as percentages, analogue to the `table_percentages`
        property. This means that the value of 3.5% will have the value 3.5 (not 0.035).
        The values can be np.nan when the corresponding percentage is also np.nan, which
        happens when the respective columns margin is 0.
        """
        return Z_975 * 100 * self.columns_std_err

    @lazyproperty
    def columns_std_dev(self):
        """standard deviation for column percentages

        `std_deviation = sqrt(variance)`
        """
        return np.sqrt(self._columns_variance)

    @lazyproperty
    def columns_std_err(self):
        """standard error for column percentages

        `std_error = sqrt(variance/N)`
        """
        return np.sqrt(self._columns_variance / self.columns_margin)

    @lazyproperty
    def counts(self):
        """2D np.float64 ndarray of weighted cube counts."""
        return np.array([row.counts for row in self._matrix.rows])

    @lazyproperty
    def cube_is_mr_aug(self):
        """True if this slice derives for an "augmented MR" cube.

        An augmented-MR cube is 3-dimensional, with the last two dimensions being the
        same MR variable. This structure allows overlap t-tests to be computed.
        """
        return self._cube.is_mr_aug

    @lazyproperty
    def description(self):
        """str description of this slice, which it takes from its rows-dimension."""
        return self._rows_dimension.description

    @lazyproperty
    def inserted_column_idxs(self):
        return tuple(
            i for i, column in enumerate(self._matrix.columns) if column.is_inserted
        )

    @lazyproperty
    def inserted_row_idxs(self):
        return tuple(i for i, row in enumerate(self._matrix.rows) if row.is_inserted)

    @lazyproperty
    def insertions(self):
        """2D np.float64 np.ma.core.MaskedArray of residuals for insertions.

              0     1     2     3      4       5       6
        0   inf   inf   inf   inf    inf    -2.9     inf
        1   inf   inf   inf   inf    inf    -4.3     inf
        2   2.5   1.3   3.3  -0.70  -7.25   -6.52    2.25
        3   inf   inf   inf   inf    inf    -2.51    inf
        4  -1.16  2.20  5.84  1.78  -8.48   -5.92    0.93
        5   inf   inf   inf   inf    inf     9.70    inf

        Only the insertions residuals are showed in a inf masked array.
        """
        inserted_rows = self.inserted_row_idxs
        inserted_cols = self.inserted_column_idxs
        if not inserted_cols and not inserted_rows:
            return []
        mask = np.zeros(self.pvals.shape)
        mask[inserted_rows, :] = 1
        mask[:, inserted_cols] = 1
        masked_pvals = np.ma.masked_array(self.pvals, np.logical_not(mask)).filled(
            np.inf
        )
        masked_zscores = np.ma.masked_array(self.zscores, np.logical_not(mask)).filled(
            np.inf
        )
        return np.stack([masked_pvals, masked_zscores])

    @lazyproperty
    def is_empty(self):
        return any(s == 0 for s in self.shape)

    @lazyproperty
    def means(self):
        return np.array([row.means for row in self._matrix.rows])

    @lazyproperty
    def min_base_size_mask(self):
        return MinBaseSizeMask(self, self._mask_size)

    @lazyproperty
    def name(self):
        """str name assigned to this slice.

        A slice takes the name of its rows-dimension.
        """
        return self.rows_dimension_name

    @lazyproperty
    def overlaps_tstats(self):
        return self._matrix.overlaps_tstats

    @lazyproperty
    def pairwise_indices(self):
        """2D ndarray of tuple of int column-idxs meeting pairwise-t threshold.

        Like::

            [
               [(1, 3, 4), (), (0,), (), ()],
               [(2,), (1, 2), (), (), (0, 3)],
               [(), (), (), (), ()],
            ]

        Has the same shape as `.counts`. Each int represents the offset of another
        column in the same row with a confidence interval meeting the threshold defined
        for this analysis.
        """
        return PairwiseSignificance.pairwise_indices(
            self, self._alpha, self._only_larger
        )

    @lazyproperty
    def pairwise_indices_alt(self):
        """2D ndarray of tuple of int column-idxs meeting alternate threshold.

        This value is None if no alternate threshold has been defined.
        """
        return (
            None
            if self._alpha_alt is None
            else PairwiseSignificance.pairwise_indices(
                self, self._alpha_alt, self._only_larger
            )
        )

    @lazyproperty
    def pairwise_significance_tests(self):
        """tuple of _ColumnPairwiseSignificance tests.

        Result has as many elements as there are columns in the slice. Each
        significance test contains `p_vals` and `t_stats` (ndarrays that represent
        probability values and statistical scores).
        """
        return tuple(
            PairwiseSignificance(self).values[column_idx]
            for column_idx in range(len(self._matrix.columns))
        )

    @lazyproperty
    def population_counts(self):
        return (
            self.table_proportions * self._population * self._cube.population_fraction
        )

    @lazyproperty
    def pvals(self):
        return np.array([row.pvals for row in self._matrix.rows])

    @lazyproperty
    def residual_test_stats(self):
        """Exposes pvals and zscores (with HS) stacked together

        Public method used as cube_method for the SOA API
        """
        return np.stack([self.pvals, self.zscores])

    @lazyproperty
    def row_base(self):
        return np.array([row.base for row in self._matrix.rows])

    @lazyproperty
    def row_labels(self):
        return tuple(row.label for row in self._matrix.rows)

    @lazyproperty
    def row_percentages(self):
        return self.row_proportions * 100

    @lazyproperty
    def row_proportions(self):
        return np.array([row.proportions for row in self._matrix.rows])

    @lazyproperty
    def rows_dimension_description(self):
        """str description assigned to rows-dimension.

        Reflects the resolved dimension-description transform cascade.
        """
        return self._rows_dimension.description

    @lazyproperty
    def rows_dimension_fills(self):
        """sequence of RGB str like "#def032" fill colors for row elements.

        The values reflect the resolved element-fill transform cascade. The length and
        ordering of the sequence correspond to the rows in the slice, including
        accounting for insertions and hidden rows.
        """
        return tuple(row.fill for row in self._matrix.rows)

    @lazyproperty
    def rows_dimension_name(self):
        """str name assigned to rows-dimension.

        Reflects the resolved dimension-name transform cascade.
        """
        return self._rows_dimension.name

    @lazyproperty
    def rows_dimension_numeric_values(self):
        return self._rows_dimension_numeric_values

    @lazyproperty
    def rows_dimension_type(self):
        """Member of `cr.cube.enum.DIMENSION_TYPE` specifying type of rows dimension."""
        return self._rows_dimension.dimension_type

    @lazyproperty
    def rows_margin(self):
        return np.array([row.margin for row in self._matrix.rows])

    @lazyproperty
    def scale_mean_pairwise_indices(self):
        """Sequence of column-idx tuples indicating pairwise-t result of scale-means.

        The calculation is based on the mean of the scale (category numeric-values) for
        each column. The length of the array is that of the columns-dimension.
        """
        return PairwiseSignificance.scale_mean_pairwise_indices(
            self, self._alpha, self._only_larger
        )

    @lazyproperty
    def scale_mean_pairwise_indices_alt(self):
        """Sequence of column-idx tuples indicating pairwise-t result of scale-means.

        Same calculation as `.scale_mean_pairwise_indices` using the `._alpha_alt`
        value. None when no secondary alpha value was specified. The length of the
        sequence is that of the columns-dimension.
        """
        if self._alpha_alt is None:
            return None

        return PairwiseSignificance.scale_mean_pairwise_indices(
            self, self._alpha_alt, self._only_larger
        )

    @lazyproperty
    def scale_means_column(self):
        """1D float64 ndarray of column scale means

        The calculation is based on multiply of the numeric values by the
        row_proportions and divide by the rows_margin.
        """
        if np.all(np.isnan(self._columns_dimension_numeric_values)):
            return None

        inner = np.nansum(
            self._columns_dimension_numeric_values * self.row_proportions, axis=1
        )
        not_a_nan_index = ~np.isnan(self._columns_dimension_numeric_values)
        denominator = np.sum(self.row_proportions[:, not_a_nan_index], axis=1)
        return inner / denominator

    @lazyproperty
    def scale_means_columns_margin(self):
        if np.all(np.isnan(self._columns_dimension_numeric_values)):
            return None

        columns_margin = self.columns_margin
        if len(columns_margin.shape) > 1:
            # Hack for MR, where column margin is a table. Figure how to
            # fix with subclassing
            columns_margin = columns_margin[0]

        not_a_nan_index = ~np.isnan(self._columns_dimension_numeric_values)
        return np.nansum(
            self._columns_dimension_numeric_values * columns_margin
        ) / np.sum(columns_margin[not_a_nan_index])

    @lazyproperty
    def scale_means_row(self):
        """1D float64 ndarray of row scale means

        The calculation is based on multiply of the numeric values by the
        column_proportions and divide by the columns_margin.
        """
        if np.all(np.isnan(self._rows_dimension_numeric_values)):
            return None
        inner = np.nansum(
            self._rows_dimension_numeric_values[:, None] * self.column_proportions,
            axis=0,
        )
        not_a_nan_index = ~np.isnan(self._rows_dimension_numeric_values)
        denominator = np.sum(self.column_proportions[not_a_nan_index, :], axis=0)
        return inner / denominator

    @lazyproperty
    def scale_means_rows_margin(self):
        if np.all(np.isnan(self._rows_dimension_numeric_values)):
            return None

        rows_margin = self.rows_margin
        if len(rows_margin.shape) > 1:
            # Hack for MR, where row margin is a table. Figure how to
            # fix with subclassing
            rows_margin = rows_margin[:, 0]

        not_a_nan_index = ~np.isnan(self._rows_dimension_numeric_values)
        return np.nansum(self._rows_dimension_numeric_values * rows_margin) / np.sum(
            rows_margin[not_a_nan_index]
        )

    @lazyproperty
    def scale_median_column(self):
        """np.int64 ndarray of the columns scale median

        The median is calculated using the standard algebra applied to the numeric
        values repeated for each related counts value
        """
        if np.all(np.isnan(self._columns_dimension_numeric_values)):
            return None
        not_a_nan_index = ~np.isnan(self._columns_dimension_numeric_values)
        numeric_values = self._columns_dimension_numeric_values[not_a_nan_index]
        counts = self.counts[:, not_a_nan_index].astype("int64")
        scale_median = np.array(
            [
                self._median(np.repeat(numeric_values, counts[i, :]))
                for i in range(counts.shape[0])
            ]
        )
        return scale_median

    @lazyproperty
    def scale_median_row(self):
        """np.int64 ndarray of the rows scale median

        The median is calculated using the standard algebra applied to the numeric
        values repeated for each related counts value
        """
        if np.all(np.isnan(self._rows_dimension_numeric_values)):
            return None
        not_a_nan_index = ~np.isnan(self._rows_dimension_numeric_values)
        numeric_values = self._rows_dimension_numeric_values[not_a_nan_index]
        counts = self.counts[not_a_nan_index, :].astype("int64")
        scale_median = np.array(
            [
                self._median(np.repeat(numeric_values, counts[:, i]))
                for i in range(counts.shape[1])
            ]
        )
        return scale_median

    @lazyproperty
    def scale_median_column_margin(self):
        """np.int64 represents the column scale median margin"""
        if np.all(np.isnan(self._columns_dimension_numeric_values)):
            return None
        columns_margin = self.columns_margin
        if len(columns_margin.shape) > 1:
            columns_margin = columns_margin[0]
        not_a_nan_index = ~np.isnan(self._columns_dimension_numeric_values)
        numeric_values = self._columns_dimension_numeric_values[not_a_nan_index]
        counts = columns_margin[not_a_nan_index].astype("int64")
        unwrapped_num_values = np.repeat(numeric_values, counts)
        return (
            np.median(unwrapped_num_values) if unwrapped_num_values.size != 0 else None
        )

    @lazyproperty
    def scale_median_row_margin(self):
        """np.int64 represents the rows scale median margin"""
        if np.all(np.isnan(self._rows_dimension_numeric_values)):
            return None
        rows_margin = self.rows_margin
        if len(rows_margin.shape) > 1:
            rows_margin = rows_margin[:, 0]
        not_a_nan_index = ~np.isnan(self._rows_dimension_numeric_values)
        numeric_values = self._rows_dimension_numeric_values[not_a_nan_index]
        counts = rows_margin[not_a_nan_index].astype("int64")
        unwrapped_num_values = np.repeat(numeric_values, counts)
        return (
            np.median(unwrapped_num_values) if unwrapped_num_values.size != 0 else None
        )

    @lazyproperty
    def scale_std_dev_column(self):
        """1D np.ndarray of the standard deviation column of scales"""
        if np.all(np.isnan(self._columns_dimension_numeric_values)):
            return None
        return np.sqrt(self.var_scale_means_column)

    @lazyproperty
    def scale_std_dev_row(self):
        """1D np.ndarray of the standard deviation row of scales"""
        if np.all(np.isnan(self._rows_dimension_numeric_values)):
            return None
        return np.sqrt(self.var_scale_means_row)

    @lazyproperty
    def scale_std_err_column(self):
        """1D np.ndarray of the standard error column of scales"""
        if np.all(np.isnan(self._columns_dimension_numeric_values)):
            return None
        return self.scale_std_dev_column / np.sqrt(self.rows_margin)

    @lazyproperty
    def scale_std_err_row(self):
        """1D np.ndarray of the standard error row of scales"""
        if np.all(np.isnan(self._rows_dimension_numeric_values)):
            return None
        return self.scale_std_dev_row / np.sqrt(self.columns_margin)

    @lazyproperty
    def shape(self):
        return self.counts.shape

    @lazyproperty
    def smoothed_dimension_dict(self):
        """dict, smoothed column dimension definition"""
        # TODO:  remove this property when the smoother gets the base measure directly
        # from the matrix later on.
        return self._columns_dimension._dimension_dict

    @lazyproperty
    def summary_pairwise_indices(self):
        return PairwiseSignificance(
            self, self._alpha, self._only_larger
        ).summary_pairwise_indices

    @lazyproperty
    def table_base(self):

        # We need to prune/order by both dimensions
        if self.dimension_types == (DT.MR, DT.MR):
            # TODO: Remove property from the assembler, when we figure out the pruning
            # by both rows and columns
            return self._matrix.table_base

        # We need to prune/order by rows
        if self.dimension_types[0] == DT.MR:
            return np.array([row.table_base for row in self._matrix.rows])

        # We need to prune/order by columns
        if self.dimension_types[1] == DT.MR:
            return np.array([column.table_base for column in self._matrix.columns])

        # No pruning or reordering since single value
        return self._matrix.table_base_unpruned

    @lazyproperty
    def table_base_unpruned(self):
        return self._matrix.table_base_unpruned

    @lazyproperty
    def table_margin(self):

        # We need to prune/order by both dimensions
        if self.dimension_types == (DT.MR, DT.MR):
            # TODO: Remove property from the assembler, when we figure out the pruning
            # by both rows and columns
            return self._matrix.table_margin

        # We need to prune/order by rows
        if self.dimension_types[0] == DT.MR:
            return np.array([row.table_margin for row in self._matrix.rows])

        # We need to prune/order by columns
        if self.dimension_types[1] == DT.MR:
            return np.array([column.table_margin for column in self._matrix.columns])

        # No pruning or reordering since single value
        return self._matrix.table_margin_unpruned

    @lazyproperty
    def table_margin_unpruned(self):
        return self._matrix.table_margin_unpruned

    @lazyproperty
    def table_name(self):
        """Provides differentiated name for each stacked table of a 3D cube."""
        if self._cube.ndim < 3:
            return None

        title = self._cube.name
        table_name = self._cube.dimensions[0].valid_elements[self._slice_idx].label

        if self._cube.is_mr_aug:
            return title

        return "%s: %s" % (title, table_name)

    @lazyproperty
    def table_percentages(self):
        return self.table_proportions * 100

    @lazyproperty
    def table_percentages_moe(self):
        """1D/2D np.float64 ndarray of margin-of-error (MoE) for table percentages.

        The values are represented as percentages, analogue to the `table_percentages`
        property. This means that the value of 3.5% will have the value 3.5 (not 0.035).
        The values can be np.nan when the corresponding percentage is also np.nan, which
        happens when the respective table margin is 0.
        """
        return Z_975 * 100 * self.table_std_err

    @lazyproperty
    def table_proportions(self):
        return np.array([row.table_proportions for row in self._matrix.rows])

    @lazyproperty
    def table_std_dev(self):
        return np.array([row.table_std_dev for row in self._matrix.rows])

    @lazyproperty
    def table_std_err(self):
        return np.array([row.table_std_err for row in self._matrix.rows])

    @lazyproperty
    def unweighted_counts(self):
        """2D tuple of unweighted counts."""
        return tuple(tuple(row.unweighted_counts) for row in self._matrix.rows)

    @lazyproperty
    def var_scale_means_column(self):
        """1D np.ndarray of the column variance values for scales

        Note: the variance for scale is defined as sum((Yi−Y~)2/(N)), where Y~ is the
              mean of the data.
        """
        if np.all(np.isnan(self._columns_dimension_numeric_values)):
            return None

        not_a_nan_index = ~np.isnan(self._columns_dimension_numeric_values)
        col_dim_numeric = self._columns_dimension_numeric_values[not_a_nan_index]

        numerator = self.counts[:, not_a_nan_index] * pow(
            np.broadcast_to(col_dim_numeric, self.counts[:, not_a_nan_index].shape)
            - self.scale_means_column.reshape(-1, 1),
            2,
        )
        denominator = np.sum(self.counts[:, not_a_nan_index], axis=1)
        return np.nansum(numerator, axis=1) / denominator

    @lazyproperty
    def var_scale_means_row(self):
        """1D np.ndarray of the row variance values for scales

        Note: the variance for scale is defined as sum((Yi−Y~)2/(N)), where Y~ is the
              mean of the data.
        """
        if np.all(np.isnan(self._rows_dimension_numeric_values)):
            return None

        not_a_nan_index = ~np.isnan(self._rows_dimension_numeric_values)
        row_dim_numeric_values = self._rows_dimension_numeric_values[not_a_nan_index]
        numerator = (
            self.counts[not_a_nan_index, :]
            * pow(
                np.broadcast_to(
                    row_dim_numeric_values, self.counts[not_a_nan_index, :].T.shape
                )
                - self.scale_means_row.reshape(-1, 1),
                2,
            ).T
        )
        denominator = np.sum(self.counts[not_a_nan_index, :], axis=0)
        return np.nansum(numerator, axis=0) / denominator

    @lazyproperty
    def zscores(self):
        return np.array([row.zscores for row in self._matrix.rows])

    # ---implementation (helpers)-------------------------------------

    @lazyproperty
    def _columns_dimension(self):
        return self._dimensions[1]

    @lazyproperty
    def _columns_dimension_numeric_values(self):
        """1D ndarray of numeric-value for each columns-dimension element."""
        return np.array([column.numeric_value for column in self._matrix.columns])

    @lazyproperty
    def _columns_variance(self):
        """variance for column percentages

        `variance = p * (1-p)`
        """
        return (
            self.counts / self.columns_margin * (1 - self.counts / self.columns_margin)
        )

    @lazyproperty
    def _dimensions(self):
        """tuple of (rows_dimension, columns_dimension) Dimension objects."""
        return tuple(
            dimension.apply_transforms(transforms)
            for dimension, transforms in zip(
                self._cube.dimensions[-2:], self._transform_dicts
            )
        )

    @lazyproperty
    def _matrix(self):
        """The TransformedMatrix object for this slice."""
        return TransformedMatrix.matrix(self._cube, self._dimensions, self._slice_idx)

    def _median(self, values):
        return np.median(values) if values.size != 0 else np.nan

    @lazyproperty
    def _rows_dimension(self):
        return self._dimensions[0]

    @lazyproperty
    def _rows_dimension_numeric_values(self):
        """1D ndarray of numeric-value for each rows-dimension element."""
        return np.array([row.numeric_value for row in self._matrix.rows])

    @lazyproperty
    def _transform_dicts(self):
        """Pair of dict (rows_dimension_transforms, columns_dimension_transforms).

        Resolved from the `transforms` argument provided on construction, it always has
        two members, even when one or both dimensions have no transforms. The transforms
        item is an empty dict (`{}`) when no transforms are specified for that
        dimension.
        """
        return (
            self._transforms_dict.get("rows_dimension", {}),
            self._transforms_dict.get("columns_dimension", {}),
        )


class _Strand(CubePartition):
    """1D cube-partition.

    A strand can arise from a 1D cube (non-CA univariate), or as a partition of
    a CA-cube (CAs are 2D) into a sequence of 1D partitions, one for each subvariable.
    """

    def __init__(self, cube, transforms, population, ca_as_0th, slice_idx, mask_size):
        super(_Strand, self).__init__(cube, transforms)
        self._population = population
        self._ca_as_0th = ca_as_0th
        self._slice_idx = slice_idx
        self._mask_size = mask_size

    @lazyproperty
    def bases(self):
        """Sequence of weighted base for each row."""
        return tuple(np.broadcast_to(self.table_margin, self.shape))

    @lazyproperty
    def counts(self):
        """tuple, 1D cube counts."""
        return tuple(row.count for row in self._stripe.rows)

    @lazyproperty
    def inserted_row_idxs(self):
        """tuple of int index of each inserted row in this strand.

        Suitable for use in applying different formatting (e.g. Bold) to inserted rows.
        Provided index values correspond to measure values as-delivered by this strand,
        after any re-ordering specified in a transform.
        """
        return tuple(i for i, row in enumerate(self._stripe.rows) if row.is_inserted)

    @lazyproperty
    def is_empty(self):
        return any(s == 0 for s in self.shape)

    @lazyproperty
    def means(self):
        return tuple(row.mean for row in self._stripe.rows)

    @lazyproperty
    def min_base_size_mask(self):
        mask = self.table_base < self._mask_size
        shape = self.shape
        return (
            mask
            if self.table_base.shape == shape
            else np.logical_or(np.zeros(shape, dtype=bool), mask)
        )

    @lazyproperty
    def name(self):
        return self.rows_dimension_name

    @lazyproperty
    def population_counts(self):
        return tuple(
            self._table_proportions_as_array
            * self._population
            * self._cube.population_fraction
        )

    @lazyproperty
    def row_base(self):
        return np.array([row.base for row in self._stripe.rows])

    @lazyproperty
    def row_count(self):
        return len(self._stripe.rows)

    @lazyproperty
    def row_labels(self):
        return tuple(row.label for row in self._stripe.rows)

    @lazyproperty
    def rows_dimension_fills(self):
        """sequence of RGB str like "#def032" fill colors for row elements.

        The values reflect the resolved element-fill transform cascade. The length and
        ordering of the sequence correspond to the rows in the slice, including
        accounting for insertions and hidden rows.
        """
        return tuple(row.fill for row in self._stripe.rows)

    @lazyproperty
    def rows_dimension_name(self):
        """str name assigned to rows-dimension.

        Reflects the resolved dimension-name transform cascade.
        """
        return self._rows_dimension.name

    @lazyproperty
    def rows_dimension_type(self):
        """Member of DIMENSION_TYPE enum describing type of rows dimension."""
        return self._rows_dimension.dimension_type

    @lazyproperty
    def rows_margin(self):
        return np.array([row.count for row in self._stripe.rows])

    @lazyproperty
    def scale_mean(self):
        """float mean of numeric-value applied to elements, or None.

        This value is `None` when no row-elements have a numeric-value assigned.
        """
        # ---return None when no row-element has been assigned a numeric value. This
        # ---avoids a division-by-zero error.
        if np.all(np.isnan(self._numeric_values)):
            return None

        # ---produce operands with rows without numeric values removed. Notably, this
        # ---excludes subtotal rows.
        numeric_values = self._numeric_values[self._numeric_values_mask]
        counts = self._counts_as_array[self._numeric_values_mask]

        # ---calculate numerator and denominator---
        total_numeric_value = np.sum(numeric_values * counts)
        total_count = np.sum(counts)

        # ---overall scale-mean is the quotient---
        return total_numeric_value / total_count

    @lazyproperty
    def scale_median(self):
        """np.int64 the median of scales

        The median is calculated using the standard algebra applied to the numeric
        values repeated for each related counts value
        """
        if np.all(np.isnan(self._numeric_values)):
            return None
        numeric_values = self._numeric_values[self._numeric_values_mask]
        counts = self._counts_as_array[self._numeric_values_mask].astype("int64")
        unwrapped_numeric_values = np.repeat(numeric_values, counts)
        return np.median(unwrapped_numeric_values)

    @lazyproperty
    def scale_std_dev(self):
        """np.float64, the standard deviation of scales"""
        if np.all(np.isnan(self._numeric_values)):
            return None
        return np.sqrt(self.var_scale_mean)

    @lazyproperty
    def scale_std_err(self):
        """np.float64, the standard error of scales"""
        if np.all(np.isnan(self._numeric_values)):
            return None
        counts = self._counts_as_array[self._numeric_values_mask]
        return np.sqrt(self.var_scale_mean / np.sum(counts))

    @lazyproperty
    def shape(self):
        """Tuple of int vector counts for this partition.

        A _Strand has a shape like (5,) which represents its row-count.

        Not to be confused with `numpy.ndarray.shape`, this represent the count of rows
        in this strand. It does not necessarily represent the shape of any underlying
        `numpy.ndarray` object In particular, the value of its row-count can be zero.
        """
        return (self.row_count,)

    @lazyproperty
    def smoothed_dimension_dict(self):
        """dict, row dimension definition"""
        return self._rows_dimension._dimension_dict

    @lazyproperty
    def standard_deviation(self):
        """np.ndarray percentages standard deviation"""
        return np.sqrt(self._variance)

    @lazyproperty
    def standard_error(self):
        """np.ndarray percentages standard error"""
        if self.dimension_types[0] == DT.MR:
            return np.sqrt(self._variance / self.bases)
        return np.sqrt(self._variance / np.sum(self.rows_margin))

    @lazyproperty
    def table_percentages_moe(self):
        """1D np.float64 ndarray of margin-of-error (MoE) for table percentages.

        The values are represented as percentages, analogue to the `table_percentages`
        property. This means that the value of 3.5% will have the value 3.5 (not 0.035).
        The values can be np.nan when the corresponding percentage is also np.nan, which
        happens when the respective columns margin is 0.
        """
        return Z_975 * 100 * self.standard_error

    @lazyproperty
    def table_base(self):
        """1D, single-element ndarray (like [3770])."""
        # For MR strands, table base is also a strand, since subvars never collapse.
        # We need to keep the ordering and hiding as in rows dimension. All this
        # information is already accessible in the underlying rows property
        # of the `_stripe`.
        if self.dimension_types[0] == DT.MR:
            return np.array([row.table_base for row in self._stripe.rows])

        # TODO: shouldn't this just be the regular value for a strand? Maybe change to
        # that if exporter always knows when it's getting this from a strand. The
        # ndarray "wrapper" seems like unnecessary baggage when we know it will always
        # be a scalar.
        return self._stripe.table_base_unpruned

    @lazyproperty
    def table_base_unpruned(self):
        return self._stripe.table_base_unpruned

    @lazyproperty
    def table_margin(self):
        # For MR strands, table base is also a strand, since subvars never collapse.
        # We need to keep the ordering and hiding as in rows dimension. All this
        # information is already accessible in the underlying rows property
        # of the `_stripe`.
        if self.dimension_types[0] == DT.MR:
            return np.array([row.table_margin for row in self._stripe.rows])

        return self._stripe.table_margin_unpruned

    @lazyproperty
    def table_margin_unpruned(self):
        return self._stripe.table_margin_unpruned

    @lazyproperty
    def table_name(self):
        """Only for CA-as-0th case, provides differentiated names for stacked tables."""
        title = self._cube.name
        table_name = self._cube.dimensions[0].valid_elements[self._slice_idx].label
        return "%s: %s" % (title, table_name)

    @lazyproperty
    def table_percentages(self):
        return tuple(self._table_proportions_as_array * 100)

    @lazyproperty
    def table_proportions(self):
        return tuple(row.table_proportions for row in self._stripe.rows)

    @lazyproperty
    def title(self):
        """The str display name of this strand, suitable for use as a column heading.

        `Strand.name` is the rows-dimension name, which is suitable for use as a title
        of the row-headings. However, a strand can also appear as a *column* and this
        value is a suitable name for such a column.
        """
        return self._cube.title

    @lazyproperty
    def unweighted_bases(self):
        """Sequence of base count for each row, before weighting.

        When the rows dimension is multiple-response, each value is different,
        reflecting the base for that individual subvariable. In all other cases, the
        table base is repeated for each row.
        """
        return tuple(np.broadcast_to(self.table_base, self.shape))

    @lazyproperty
    def unweighted_counts(self):
        """tuple of int unweighted count for each row of stripe."""
        return tuple(row.unweighted_count for row in self._stripe.rows)

    @lazyproperty
    def var_scale_mean(self):
        if np.all(np.isnan(self._numeric_values)):
            return None

        numeric_values = self._numeric_values[self._numeric_values_mask]
        counts = self._counts_as_array[self._numeric_values_mask]

        return np.nansum(counts * pow((numeric_values - self.scale_mean), 2)) / np.sum(
            counts
        )

    # ---implementation (helpers)-------------------------------------

    @lazyproperty
    def _counts_as_array(self):
        """1D ndarray of count for each row."""
        return np.array([row.count for row in self._stripe.rows_including_hidden])

    @lazyproperty
    def _dimensions(self):
        """tuple of (row,) Dimension object."""
        return (self._rows_dimension,)

    @lazyproperty
    def _numeric_values(self):
        """Array of numeric-value for each element in rows dimension.

        The items in the array can be numeric or np.nan, which appears for an inserted
        row (subtotal) or where the row-element has been assigned no numeric value.
        """
        return np.array(
            [row.numeric_value for row in self._stripe.rows_including_hidden]
        )

    @lazyproperty
    def _numeric_values_mask(self):
        """np.ndarray boolean elements for each element in rows dimension."

        This array contains True or False according to the nan in the numeric_values
        array
        """
        is_a_number_mask = ~np.isnan(self._numeric_values)
        return is_a_number_mask

    @lazyproperty
    def _rows_dimension(self):
        """Dimension object for the single dimension of this strand."""
        return self._cube.dimensions[-1].apply_transforms(self._row_transforms_dict)

    @lazyproperty
    def _row_transforms_dict(self):
        """Transforms dict for the single (rows) dimension of this strand."""
        return self._transforms_dict.get("rows_dimension", {})

    @lazyproperty
    def _stripe(self):
        """The post-transforms 1D data-partition for this strand."""
        return TransformedStripe.stripe(
            self._cube, self._rows_dimension, self._ca_as_0th, self._slice_idx
        )

    @lazyproperty
    def _table_proportions_as_array(self):
        return np.array([row.table_proportions for row in self._stripe.rows])

    @lazyproperty
    def _variance(self):
        """variance for cell percentages

        `variance = p * (1-p)`
        """
        p = self._table_proportions_as_array
        return p * (1 - p)


class _Nub(CubePartition):
    """0D slice."""

    @lazyproperty
    def is_empty(self):
        return False if self.unweighted_count else True

    @lazyproperty
    def means(self):
        return self._scalar.means

    @lazyproperty
    def table_base(self):
        return self._scalar.table_base

    @lazyproperty
    def unweighted_count(self):
        return self._cube.unweighted_counts

    # ---implementation (helpers)-------------------------------------

    @lazyproperty
    def _dimensions(self):
        return ()

    @lazyproperty
    def _scalar(self):
        """The pre-transforms data-array for this slice."""
        return MeansScalar(self._cube.counts, self._cube.unweighted_counts)



# encoding: utf-8

"""This module contains tests for the correct index functionality."""

import numpy as np

from cr.cube.crunch_cube import CrunchCube

from ...fixtures import CR


def test_cat_x_cat_slice_column_index():
    cube = CrunchCube(CR.CAT_X_CAT)
    expected = np.array([[100, 80], [100, 120]])
    actual = cube.slices[0].index_table(axis=0)
    np.testing.assert_almost_equal(actual, expected)


def test_cat_x_cat_slice_row_index():
    cube = CrunchCube(CR.CAT_X_CAT)
    expected = np.array([[107.142857142857, 85.7142857142857], [93.75, 112.5]])
    actual = cube.slices[0].index_table(axis=1)
    np.testing.assert_almost_equal(actual, expected)


def test_cat_x_cat_slice_row_index_with_baseline():
    cube = CrunchCube(CR.CAT_X_CAT)
    expected = np.array(
        [[119.047619047619, 71.4285714285714], [104.16666666666667, 93.75]]
    )
    actual = cube.slices[0].index_table(axis=1, baseline=np.array([0.6, 0.4]))
    np.testing.assert_almost_equal(actual, expected)


def test_cat_x_cat_slice_column_index_with_baseline():
    cube = CrunchCube(CR.CAT_X_CAT)
    expected = np.array([[83.3333333333333, 66.6666666666667], [125, 150]])
    actual = cube.slices[0].index_table(axis=0, baseline=np.array([0.6, 0.4]))
    np.testing.assert_almost_equal(actual, expected)


def test_mr_x_cat_slice_column_index():
    cube = CrunchCube(CR.SELECTED_CROSSTAB_4)
    expected = np.array(
        [
            [95.8651525855387, 103.859044435659],
            [102.305106635277, 97.8432045727022],
            [97.6031146323114, 102.24274029149],
            [98.1029444304978, 101.829068203842],
            [114.466510106092, 86.0656684647625],
            [99.2925720053358, 100.682933745397],
        ]
    )
    actual = cube.slices[0].index_table(axis=0)
    np.testing.assert_almost_equal(actual, expected)


def test_mr_x_cat_slice_row_index():
    cube = CrunchCube(CR.SELECTED_CROSSTAB_4)
    expected = np.array(
        [
            [95.3416155822363, 104.394053238208],
            [101.879419344372, 98.2272247381305],
            [97.1985863211465, 102.642452778304],
            [99.2098729346168, 100.745292805163],
            [115.700063998356, 85.1908063256891],
            [100.477252947149, 99.5498278652431],
        ]
    )
    actual = cube.slices[0].index_table(axis=1)
    np.testing.assert_almost_equal(actual, expected)


def test_mr_x_mr_slice_column_index():
    mr_x_mr = CrunchCube(CR.FULL_CUBE)
    mr_alone = CrunchCube(CR.MR_WGTD)
    expected = mr_x_mr.proportions(axis=0) / mr_alone.proportions()[:, None] * 100
    actual = mr_x_mr.slices[0].index_table(axis=0)
    np.testing.assert_almost_equal(actual, expected)


def test_mr_x_mr_slice_row_index():
    mr_x_mr = CrunchCube(CR.FULL_CUBE)
    expected = np.array(
        [
            96.5399786,
            101.08725891,
            89.22034634,
            89.31908705,
            97.03625304,
            99.96205366,
            79.74421663,
            94.32481007,
            98.95581177,
            101.14422043,
            95.50739913,
            91.82091585,
            92.66856944,
            102.10571127,
            94.88279498,
            84.22596655,
            92.62629722,
            86.77934972,
            99.31115914,
            98.72846269,
            99.54678433,
            94.13302782,
            101.99733805,
            102.24392708,
            97.87112979,
            95.08750269,
            100.61288629,
        ]
    )
    actual = mr_x_mr.slices[0].index_table(axis=1)[0]
    np.testing.assert_almost_equal(actual, expected)


def test_cat_x_mr_slice_column_index():
    cat_x_mr = CrunchCube(CR.CAT_X_MR)
    expected = np.array(
        [[90.9090909, 106.9518717, 95.6937799], [104.4776119, 96.5759438, 102.1209741]]
    )
    actual = cat_x_mr.slices[0].index_table(axis=0)
    np.testing.assert_almost_equal(actual, expected)


def test_cat_x_mr_slice_row_index():
    cat_x_mr = CrunchCube(CR.CAT_X_MR)
    expected = np.array(
        [[85.7142857, 111.5294118, 96.1098398], [107.6923077, 94.6623094, 101.9036954]]
    )
    actual = cat_x_mr.slices[0].index_table(axis=1)
    np.testing.assert_almost_equal(actual, expected)


def test_mr_x_mr_index_tables_parity_with_whaam_and_r():
    cat_x_mr = CrunchCube(CR.MR_X_MR_INDEX_TABLE)
    # Test column direction
    expected = np.array(
        [
            [192.05298013, 97.23165321, 89.68799602],
            [99.22588537, 239.38592924, 95.99096915],
            [93.52597694, 98.08689727, 182.31556654],
        ]
    )
    actual = cat_x_mr.slices[0].index_table(axis=0)
    np.testing.assert_almost_equal(actual, expected)
    # Test row direction
    expected = np.array(
        [
            [192.0529801, 99.2258854, 93.5259769],
            [97.2316532, 239.3859292, 98.0868973],
            [89.687996, 95.9909692, 182.3155665],
        ]
    )
    actual = cat_x_mr.slices[0].index_table(axis=1)
    np.testing.assert_almost_equal(actual, expected)


def test_mr_x_3vl_index_tables_parity_with_nssat():
    mr_x_3vl = CrunchCube(CR.NSSAT_MR_X_3vl)
    # Test column direction
    expected = np.array(
        [
            [179.793686976007, 90.4924459426829],
            [182.343952039497, 88.5838171105893],
            [84.6174957937067, 101.993855386627],
            [np.nan, np.nan],
        ]
    )
    actual = mr_x_3vl.slices[0].index_table(axis=0)
    np.testing.assert_almost_equal(actual, expected)
    # Test row direction
    expected = np.array(
        [
            [179.360970521192, 90.2746542556969],
            [184.52243729878, 89.6421386925666],
            [84.5355650535557, 101.895100017134],
            [np.nan, np.nan],
        ]
    )
    actual = mr_x_3vl.slices[0].index_table(axis=1)
    np.testing.assert_almost_equal(actual, expected)


def test_mr_x_mr_index_tables_parity_with_nssat():
    mr_x_mr = CrunchCube(CR.NSSAT_MR_X_MR)
    # Test column direction
    expected = np.array(
        [
            [
                114.917891097666,
                94.6007480891202,
                75.7981149285497,
                41.5084915084915,
                64.5687645687646,
                581.118881118881,
                np.nan,
                0,
                np.nan,
            ],
            [
                90.0597657183839,
                95.9426026719446,
                102.497687326549,
                84.1945288753799,
                261.93853427896,
                0,
                np.nan,
                0,
                np.nan,
            ],
            [
                99.4879510762734,
                101.567130443518,
                101.446145177951,
                106.834310398025,
                86.4170866330693,
                59.8272138228942,
                np.nan,
                119.654427645788,
                np.nan,
            ],
            [np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan],
        ]
    )
    actual = mr_x_mr.slices[0].index_table(axis=0)
    np.testing.assert_almost_equal(actual, expected)
    # Test row direction
    expected = np.array(
        [
            [
                104.349919743178,
                85.9011627906977,
                68.8276397515528,
                37.6913265306122,
                58.6309523809524,
                527.678571428571,
                np.nan,
                0,
                np.nan,
            ],
            [
                98.1631656082071,
                104.575328614762,
                111.7202268431,
                91.7701863354037,
                285.507246376812,
                0,
                np.nan,
                0,
                np.nan,
            ],
            [
                99.6740889304191,
                101.757158356526,
                101.635946732516,
                107.03419298754,
                86.57876943881,
                59.9391480730223,
                np.nan,
                119.878296146045,
                np.nan,
            ],
            [np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan],
        ]
    )
    actual = mr_x_mr.slices[0].index_table(axis=1)
    np.testing.assert_almost_equal(actual, expected)


def test_mr_single_cat_x_mr():
    cube_slice = CrunchCube(CR.MR_SINGLE_CAT_X_MR).slices[0]
    expected = np.array([[100, 100, np.nan]])
    np.testing.assert_array_equal(cube_slice.index_table(axis=0), expected)
    np.testing.assert_array_equal(cube_slice.index_table(axis=1), expected)

    # With pruning
    expected = np.ma.masked_array([[100, 100, np.nan]], mask=[[False, False, True]])
    np.testing.assert_array_equal(cube_slice.index_table(axis=0, prune=True), expected)


def test_mr_x_mr_single_cat():
    cube_slice = CrunchCube(CR.MR_X_MR_SINGLE_CAT).slices[0]
    expected = np.array([100, 100, np.nan])
    np.testing.assert_array_equal(cube_slice.index_table(axis=0), expected)

    # With pruning
    expected = np.ma.masked_array([100, 100, np.nan], mask=[False, False, True])
    np.testing.assert_array_equal(cube_slice.index_table(axis=0, prune=True), expected)



# encoding: utf-8

"""Integration test suite for the cr.cube.legacy_dimension module.

This should go away when the legacy CrunchCube is dropped.
"""

from __future__ import absolute_import, division, print_function, unicode_literals

import numpy as np
import pytest

from unittest import TestCase

from cr.cube.crunch_cube import CrunchCube
from cr.cube.legacy_dimension import AllDimensions, Dimension, _Subtotal
from cr.cube.enums import DIMENSION_TYPE as DT

from ...fixtures import CR  # ---mnemonic: CR = 'cube-response'---


class DescribeIntegratedAllDimensions(object):
    """Integration-test suite for `cr.cube.legacy_dimension.AllDimensions` object."""

    @pytest.mark.parametrize(
        "cube_response, expected_types",
        (
            (CR.CAT_X_CAT, (DT.CAT, DT.CAT)),
            (CR.CA_X_MR_WEIGHTED_HS, (DT.CA, DT.CA_CAT, DT.MR, DT.MR_CAT)),
        ),
    )
    def it_resolves_the_type_of_each_dimension(self, cube_response, expected_types):
        dimension_dicts = cube_response["result"]["dimensions"]
        all_dimensions = AllDimensions(dimension_dicts)

        dimension_types = tuple(d.dimension_type for d in all_dimensions)

        assert dimension_types == expected_types

    def it_provides_access_to_the_apparent_dimensions(self):
        dimension_dicts = CR.CA_X_MR_WEIGHTED_HS["result"]["dimensions"]
        all_dimensions = AllDimensions(dimension_dicts)

        apparent_dimension_types = tuple(
            d.dimension_type for d in all_dimensions.apparent_dimensions
        )

        assert apparent_dimension_types == (DT.CA, DT.CA_CAT, DT.MR)


class TestDimension(TestCase):
    """Legacy integration-test suite for Dimension object."""

    def test_subtotals_indices_single_subtotal(self):
        dimension = CrunchCube(CR.ECON_BLAME_WITH_HS).dimensions[0]
        hs_indices = dimension.hs_indices
        self.assertEqual(hs_indices, ((1, (0, 1)),))

    def test_inserted_hs_indices_single_subtotal(self):
        dimension = CrunchCube(CR.ECON_BLAME_WITH_HS).dimensions[0]
        # It can be verified what the inserted indices are, by comparing
        # labels with/without transforms.
        expected = [2]
        actual = dimension.inserted_hs_indices
        assert actual == expected

    def test_labels_for_categoricals(self):
        dimension_dict = {
            "type": {
                "class": "categorical",
                "categories": [
                    {"id": 1, "name": "Cat", "missing": False},
                    {"id": 2, "name": "Mouse", "missing": False},
                    {"id": -1, "name": "Iguana", "missing": True},
                ],
            }
        }
        dimension = Dimension(dimension_dict, DT.CAT)

        # ---get only non-missing---
        labels = dimension.labels()
        assert labels == ["Cat", "Mouse"]

        # ---get all---
        labels = dimension.labels(include_missing=True)
        assert labels == ["Cat", "Mouse", "Iguana"]

    def test_labels_for_numericals(self):
        dimension_dict = {
            "type": {
                "class": "enum",
                "elements": [
                    {"id": 0, "value": "smallish", "missing": False},
                    {"id": 1, "value": "kinda big", "missing": False},
                    {"id": 2, "value": {}, "missing": True},
                ],
                "subtype": {"class": "numeric"},
            }
        }
        dimension = Dimension(dimension_dict, DT.BINNED_NUMERIC)

        # ---non-missing labels---
        labels = dimension.labels()
        assert labels == ["smallish", "kinda big"]

        # ---all labels, both valid and missing---
        labels = dimension.labels(include_missing=True)
        assert labels == ["smallish", "kinda big", ""]

        # ---all labels, both valid and missing---
        labels = dimension.labels(include_cat_ids=True)
        assert labels == [("smallish", 0), ("kinda big", 1)]

    def test_subtotals_indices_two_subtotals(self):
        dimension = CrunchCube(CR.ECON_BLAME_WITH_HS_MISSING).dimensions[0]
        hs_indices = dimension.hs_indices
        self.assertEqual(hs_indices, ((1, (0, 1)), ("bottom", (3, 4))))

    def test_inserted_hs_indices_two_subtotals(self):
        dimension = CrunchCube(CR.ECON_BLAME_WITH_HS_MISSING).dimensions[0]
        # It can be verified what the inserted indices are, by comparing
        # labels with/without transforms.
        expected = [2, 6]
        actual = dimension.inserted_hs_indices
        self.assertEqual(actual, expected)

    def test_inserted_hs_indices_order_and_labels(self):
        dimension_dict = {
            "references": {
                "view": {
                    "transform": {
                        "insertions": [
                            {
                                "anchor": "bottom",
                                "args": [111],
                                "function": "subtotal",
                                "name": "bottoms up one",
                            },
                            {
                                "anchor": "bottom",
                                "args": [222],
                                "function": "subtotal",
                                "name": "bottoms up two",
                            },
                            {
                                "anchor": "bottom",
                                "args": [333],
                                "function": "subtotal",
                                "name": "bottoms up three",
                            },
                            {
                                "anchor": "top",
                                "args": [444],
                                "function": "subtotal",
                                "name": "on top one",
                            },
                            {
                                "anchor": "top",
                                "args": [555],
                                "function": "subtotal",
                                "name": "on top two",
                            },
                            {
                                "anchor": 333,
                                "args": [555],
                                "function": "subtotal",
                                "name": "in the middle one",
                            },
                            {
                                "anchor": 333,
                                "args": [555],
                                "function": "subtotal",
                                "name": "in the middle two",
                            },
                        ]
                    }
                }
            },
            "type": {
                "categories": [
                    {"id": 111},
                    {"id": 222},
                    {"id": 333},
                    {"id": 444},
                    {"id": 555},
                ],
                "class": "categorical",
            },
        }
        dimension = Dimension(dimension_dict, DT.CAT)

        assert dimension.inserted_hs_indices == [0, 1, 5, 6, 9, 10, 11]
        assert dimension.labels(include_transforms=True) == [
            "on top one",
            "on top two",
            "",
            "",
            "",
            "in the middle one",
            "in the middle two",
            "",
            "",
            "bottoms up one",
            "bottoms up two",
            "bottoms up three",
        ]

    def test_has_transforms_false(self):
        dimension = CrunchCube(CR.ECON_BLAME_X_IDEOLOGY_ROW_HS).dimensions[1]
        expected = False
        actual = dimension.has_transforms
        self.assertEqual(actual, expected)

    def test_has_transforms_true(self):
        dimension = CrunchCube(CR.ECON_BLAME_X_IDEOLOGY_ROW_HS).dimensions[0]
        expected = True
        actual = dimension.has_transforms
        self.assertEqual(actual, expected)

    def test_hs_indices_for_mr(self):
        dimension = CrunchCube(CR.MR_X_CAT_HS)._all_dimensions[1]
        hs_indices = dimension.hs_indices
        assert hs_indices == ()

    def test_hs_indices_with_bad_data(self):
        cube = CrunchCube(CR.CA_WITH_NETS)

        subvar_dim = cube.dimensions[0]
        anchor_idxs = [anchor_idx for anchor_idx, _ in subvar_dim.hs_indices]
        assert anchor_idxs == ["bottom", "bottom"]

        cat_dim = cube.dimensions[1]
        anchor_idxs = [anchor_idx for anchor_idx, _ in cat_dim.hs_indices]
        assert anchor_idxs == ["bottom", "bottom"]

    def test_skips_bad_data_for_hs_indices(self):
        """Test H&S indices with bad input data.

        This test ensures that H&S functionality doesn't break if it
        encounters bad transformations data, as is possible with some of the
        leftovers in the variables.
        """
        dimension_dict = {
            "references": {
                "view": {
                    "transform": {
                        "insertions": [
                            {"anchor": 101, "name": "This is respondent ideology"},
                            {
                                "anchor": 2,
                                "args": [1, 2],
                                "function": "subtotal",
                                "name": "Liberal net",
                            },
                            {
                                "anchor": 5,
                                "args": [5, 4],
                                "function": "subtotal",
                                "name": "Conservative net",
                            },
                            {
                                "anchor": "fake anchor",
                                "args": ["fake_arg_1", "fake_arg_2"],
                                "function": "fake_fcn_name_not_subtotal",
                                "name": "Fake Name",
                            },
                        ]
                    }
                }
            },
            "type": {
                "categories": [
                    {
                        "numeric_value": 1,
                        "id": 1,
                        "name": "President Obama",
                        "missing": False,
                    },
                    {
                        "numeric_value": 2,
                        "id": 2,
                        "name": "Republicans in Congress",
                        "missing": False,
                    },
                    {"numeric_value": 5, "id": 5, "name": "Not sure", "missing": False},
                    {"numeric_value": 4, "id": 4, "name": "Neither", "missing": False},
                ],
                "class": "categorical",
                "ordinal": False,
            },
        }
        dimension = Dimension(dimension_dict, DT.CAT)

        hs_indices = dimension.hs_indices

        assert hs_indices == ((1, (0, 1)), (2, (2, 3)))

    def test_subtotals(self):
        dimension_dict = {
            "references": {
                "view": {
                    "transform": {
                        "insertions": [
                            {"anchor": 101, "name": "This is respondent ideology"},
                            {
                                "anchor": 2,
                                "args": [1, 2],
                                "function": "subtotal",
                                "name": "Liberal net",
                            },
                            {
                                "anchor": 5,
                                "args": [5, 4],
                                "function": "subtotal",
                                "name": "Conservative net",
                            },
                            {
                                "anchor": "fake anchor",
                                "args": ["fake_arg_1", "fake_arg_2"],
                                "function": "fake_fcn_name_not_subtotal",
                                "name": "Fake Name",
                            },
                        ]
                    }
                }
            },
            "type": {
                "categories": [{"id": 1}, {"id": 5}, {"id": 8}, {"id": 9}, {"id": -1}],
                "class": "categorical",
            },
        }
        dimension = Dimension(dimension_dict, DT.CAT)

        subtotals = dimension._subtotals

        assert len(subtotals) == 2

        subtotal = subtotals[0]
        assert isinstance(subtotal, _Subtotal)
        assert subtotal.anchor == "bottom"
        assert subtotal.addend_ids == (1,)
        assert subtotal.addend_idxs == (0,)
        assert subtotal.label == "Liberal net"

        subtotal = subtotals[1]
        assert isinstance(subtotal, _Subtotal)
        assert subtotal.anchor == 5
        assert subtotal.addend_ids == (5,)
        assert subtotal.addend_idxs == (1,)
        assert subtotal.label == "Conservative net"

    def test_numeric_values(self):
        dimension_dict = {
            "type": {
                "categories": [
                    {"id": 42, "missing": False, "numeric_value": 1},
                    {"id": 43, "missing": False, "numeric_value": 2},
                    {"id": 44, "missing": True, "numeric_value": 3},
                    {"id": 45, "missing": False, "numeric_value": None},
                    {"id": 46, "missing": False},
                ],
                "class": "categorical",
            }
        }
        dimension = Dimension(dimension_dict, DT.CAT)

        numeric_values = dimension.numeric_values

        self.assertEqual(numeric_values, (1, 2, np.nan, np.nan))



# encoding: utf-8

"""Unit test suite for `cr.cube.stripe` module."""

from __future__ import absolute_import, division, print_function, unicode_literals

import pytest

from cr.cube.cube import Cube
from cr.cube.dimension import Dimension, _Element, _Subtotal
from cr.cube.enums import DIMENSION_TYPE as DT
from cr.cube.stripe import (
    _BaseBaseStripe,
    _BaseStripeRow,
    _MeansStripeRow,
    _StripeInsertedRow,
    _StripeInsertionHelper,
    TransformedStripe,
)

from ..unitutil import (
    ANY,
    call,
    class_mock,
    initializer_mock,
    instance_mock,
    method_mock,
    property_mock,
)


class DescribeTransformedStripe(object):
    """Unit test suite for `cr.cube.stripe.TransformedStripe` object."""

    def it_provides_a_constructor_classmethod(
        self, request, cube_, dimension_, base_stripe_, stripe_
    ):
        _BaseBaseStripe_ = class_mock(request, "cr.cube.stripe._BaseBaseStripe")
        _BaseBaseStripe_.factory.return_value = base_stripe_
        _init_ = initializer_mock(request, TransformedStripe)

        stripe = TransformedStripe.stripe(cube_, dimension_, True, 42)

        _BaseBaseStripe_.factory.assert_called_once_with(cube_, dimension_, True, 42)
        _init_.assert_called_once_with(ANY, dimension_, base_stripe_)
        assert isinstance(stripe, TransformedStripe)

    def it_provides_access_to_its_rows(self, request):
        rows_ = tuple(
            instance_mock(
                request, _BaseStripeRow, hidden=bool(i % 2), name="row[%d]" % i
            )
            for i in range(4)
        )
        property_mock(
            request, TransformedStripe, "rows_including_hidden", return_value=rows_
        )
        stripe = TransformedStripe(None, None)

        assert stripe.rows == (rows_[0], rows_[2])

    def it_provides_access_to_its_rows_including_hidden_ones(self, request, dimension_):
        rows_ = tuple(
            instance_mock(
                request, _BaseStripeRow, name="row[%d]" % i, hidden=bool(i % 2)
            )
            for i in range(4)
        )
        property_mock(
            request,
            TransformedStripe,
            "_ordered_rows",
            return_value=("ordered", "rows"),
        )
        property_mock(request, TransformedStripe, "_table_margin", return_value=42)
        _StripeInsertionHelper_ = class_mock(
            request, "cr.cube.stripe._StripeInsertionHelper"
        )
        _StripeInsertionHelper_.iter_interleaved_rows.return_value = iter(rows_)
        stripe = TransformedStripe(dimension_, None)

        rows = stripe.rows_including_hidden

        _StripeInsertionHelper_.iter_interleaved_rows.assert_called_once_with(
            dimension_, ("ordered", "rows"), 42
        )
        assert rows == (rows_[0], rows_[1], rows_[2], rows_[3])

    @pytest.mark.parametrize(
        "display_order", ((), (0,), (1, 0), (0, 1, 2), (2, 1, 0, 3))
    )
    def it_provides_access_to_the_ordered_base_rows_to_help(
        self, request, display_order, dimension_, base_stripe_
    ):
        dimension_.display_order = display_order
        rows_ = tuple(
            instance_mock(request, _BaseStripeRow, name="row[%d]" % i)
            for i in range(len(display_order))
        )
        base_stripe_.rows = rows_
        stripe = TransformedStripe(dimension_, base_stripe_)

        ordered_rows = stripe._ordered_rows

        assert ordered_rows == tuple(rows_[idx] for idx in display_order)

    # fixture components ---------------------------------------------

    @pytest.fixture
    def base_stripe_(self, request):
        return instance_mock(request, _BaseBaseStripe)

    @pytest.fixture
    def cube_(self, request):
        return instance_mock(request, Cube)

    @pytest.fixture
    def dimension_(self, request):
        return instance_mock(request, Dimension)

    @pytest.fixture
    def stripe_(self, request):
        return instance_mock(request, TransformedStripe)


class Describe_StripeInsertionHelper(object):
    """Unit test suite for `cr.cube.stripe._StripeInsertionHelper` function-object."""

    def it_provides_an_interface_classmethod(self, request, dimension_):
        _init_ = initializer_mock(request, _StripeInsertionHelper)
        _iter_interleaved_rows_ = method_mock(
            request,
            _StripeInsertionHelper,
            "_iter_interleaved_rows",
            return_value=iter(("interleaved", "rows")),
        )

        interleaved_rows = _StripeInsertionHelper.iter_interleaved_rows(
            dimension_, ("ordered", "rows"), 42
        )

        _init_.assert_called_once_with(ANY, dimension_, ("ordered", "rows"), 42)
        _iter_interleaved_rows_.assert_called_once_with(ANY)
        assert next(interleaved_rows) == "interleaved"
        assert next(interleaved_rows) == "rows"
        assert next(interleaved_rows, None) is None

    def it_generates_the_interleaved_rows_in_order_to_help(self, request):
        subtotal_rows_ = (
            instance_mock(request, _StripeInsertedRow, name="subt_0", anchor=1),
            instance_mock(request, _StripeInsertedRow, name="subt_1", anchor="bottom"),
            instance_mock(request, _StripeInsertedRow, name="subt_2", anchor=2),
            instance_mock(request, _StripeInsertedRow, name="subt_3", anchor="top"),
            instance_mock(request, _StripeInsertedRow, name="subt_4", anchor="top"),
        )
        property_mock(
            request,
            _StripeInsertionHelper,
            "_inserted_rows",
            return_value=subtotal_rows_,
        )
        ordered_rows_ = (
            instance_mock(request, _BaseStripeRow, name="row_0", element_id=1),
            instance_mock(request, _BaseStripeRow, name="row_1", element_id=2),
        )
        insertion_helper = _StripeInsertionHelper(None, ordered_rows_, None)

        interleaved_rows = insertion_helper._iter_interleaved_rows()

        assert next(interleaved_rows) == subtotal_rows_[3]
        assert next(interleaved_rows) == subtotal_rows_[4]
        assert next(interleaved_rows) == ordered_rows_[0]
        assert next(interleaved_rows) == subtotal_rows_[0]
        assert next(interleaved_rows) == ordered_rows_[1]
        assert next(interleaved_rows) == subtotal_rows_[2]
        assert next(interleaved_rows) == subtotal_rows_[1]
        assert next(interleaved_rows, None) is None

    @pytest.mark.parametrize(
        ("dimension_type", "expected_len"), ((DT.MR, 0), (DT.CA, 0), (DT.CAT, 3))
    )
    def it_assembles_the_inserted_rows_to_help(
        self, request, dimension_type, expected_len, dimension_
    ):
        subtotal_rows_ = (
            instance_mock(request, _StripeInsertedRow, name="subtotal_0"),
            instance_mock(request, _StripeInsertedRow, name="subtotal_1"),
            instance_mock(request, _StripeInsertedRow, name="subtotal_2"),
        )
        _StripeInsertedRow_ = class_mock(
            request, "cr.cube.stripe._StripeInsertedRow", side_effect=subtotal_rows_
        )
        dimension_.dimension_type = dimension_type
        dimension_.subtotals = subtotal_rows_
        insertion_helper = _StripeInsertionHelper(dimension_, ("ordered", "rows"), 42)

        inserted_rows = insertion_helper._inserted_rows

        assert _StripeInsertedRow_.call_args_list == [
            call(row, ("ordered", "rows"), 42) for row in subtotal_rows_[:expected_len]
        ]
        assert inserted_rows == subtotal_rows_[:expected_len]

    # fixture components ---------------------------------------------

    @pytest.fixture
    def dimension_(self, request):
        return instance_mock(request, Dimension)


class Describe_StripeInsertedRow(object):
    """Unit test suite for `cr.cube.stripe._StripeInsertedRow` object."""

    @pytest.mark.parametrize("anchor_value", ("top", "bottom", 1, 36))
    def it_knows_its_anchor_location(self, anchor_value, subtotal_):
        subtotal_.anchor = anchor_value
        inserted_row = _StripeInsertedRow(subtotal_, None, None)

        anchor = inserted_row.anchor

        assert anchor == anchor_value

    def it_knows_it_is_inserted(self):
        assert _StripeInsertedRow(None, None, None).is_inserted is True

    def it_knows_its_unweighted_count(self, request):
        addend_rows_ = tuple(
            instance_mock(request, _BaseStripeRow, unweighted_count=i + 1)
            for i in range(3)
        )
        property_mock(
            request, _StripeInsertedRow, "_addend_rows", return_value=addend_rows_
        )
        inserted_row = _StripeInsertedRow(None, None, None)

        unweighted_count = inserted_row.unweighted_count

        assert unweighted_count == 6  # --- 1+2+3 ---

    def it_gathers_the_addend_rows_to_help(self, request, subtotal_):
        base_rows_ = tuple(
            instance_mock(
                request, _BaseStripeRow, name="base_rows_[%d]" % i, element_id=i + 1
            )
            for i in range(4)
        )
        subtotal_.addend_ids = (1, 3)
        inserted_row = _StripeInsertedRow(subtotal_, base_rows_, None)

        addend_rows = inserted_row._addend_rows

        assert addend_rows == (base_rows_[0], base_rows_[2])

    # fixture components ---------------------------------------------

    @pytest.fixture
    def subtotal_(self, request):
        return instance_mock(request, _Subtotal)


class Describe_BaseStripeRow(object):
    """Unit test suite for `cr.cube.stripe._BaseStripeRow` object."""

    def it_knows_its_element_id(self, element_):
        element_.element_id = 42
        row = _BaseStripeRow(element_)

        element_id = row.element_id

        assert element_id == 42

    def it_knows_it_is_not_inserted(self):
        assert _BaseStripeRow(None).is_inserted is False

    # fixture components ---------------------------------------------

    @pytest.fixture
    def element_(self, request):
        return instance_mock(request, _Element)


class Describe_MeansStripeRow(object):
    """Unit test suite for `cr.cube.stripe.Describe_MeansStripeRow` object."""

    def it_knows_its_unweighted_count(self):
        assert _MeansStripeRow(None, 42, None).unweighted_count == 42



# encoding: utf-8

"""JSON cube-response source files for testing purposes."""

import json
import os

CUBES_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), "cubes")


class LazyCubeResponseLoader(object):
    """Loads and caches cube-responses by name from fixture directory.

    Provides access to all the cube-response fixtures in a directory by
    a standardized mapping of the file name, e.g. cat-x-cat.json is available
    as the `.CAT_X_CAT` attribute of the loader.

    The fixture directory is specified relative to this (fixture root)
    directory.
    """

    def __init__(self, relpath):
        self._relpath = relpath
        self._cache = {}

    def __getattr__(self, fixture_name):
        """Return cube-dict from JSON file mapping to *fixture_name*.

        A *fixture_name* like 'CAT_X_CAT' will map to the JSON file
        'cat-x-cat.json' in the directory specified on construction.
        """
        if fixture_name not in self._cache:
            self._load_to_cache(fixture_name)
        return self._cache[fixture_name]

    @property
    def _dirpath(self):
        """Absolute path to relative directory specified in relpath."""
        thisdir = os.path.dirname(os.path.abspath(__file__))
        return os.path.abspath(os.path.join(thisdir, self._relpath))

    def _json_path(self, fixture_name):
        """Return absolute path to JSON file for *fixture_name*."""
        return "%s/%s.json" % (self._dirpath, fixture_name.replace("_", "-").lower())

    def _load_json(self, path):
        """Return dict parsed from JSON at *path*."""
        with open(path) as f:
            cube_response = json.load(f)
        return cube_response

    def _load_to_cache(self, fixture_name):
        json_path = self._json_path(fixture_name)
        if not os.path.exists(json_path):
            raise ValueError("no JSON fixture found at %s" % json_path)
        self._cache[fixture_name] = self._load_json(json_path)


CR = LazyCubeResponseLoader(".")  # ---mnemonic: CR = 'cube-response'---
SM = LazyCubeResponseLoader("./scale_means")
TR = LazyCubeResponseLoader("./transforms")



# encoding: utf-8

import numpy as np

from unittest import TestCase

from cr.cube.crunch_cube import CrunchCube
from cr.cube.measures.index import Index

from ...fixtures import CR


class TestIndex(TestCase):
    def test_mr_x_cat_index(self):
        cube = CrunchCube(CR.SELECTED_CROSSTAB_4)
        expected = np.array(
            [
                [0.95865152585539, 1.0385904443566],
                [1.02305106635277, 0.9784320457270],
                [0.97603114632311, 1.0224274029149],
                [0.98102944430498, 1.0182906820384],
                [1.14466510106092, 0.8606566846476],
                [0.99292572005336, 1.0068293374540],
            ]
        )
        actual = Index.data(cube, weighted=True, prune=False)
        np.testing.assert_almost_equal(actual, expected)

    def test_cat_x_mr_index(self):
        cube = CrunchCube(CR.CAT_X_MR)
        expected = np.array(
            [[0.8571429, 1.1152941, 0.9610984], [1.0769231, 0.9466231, 1.019037]]
        )
        actual = Index.data(cube, weighted=True, prune=False)
        np.testing.assert_almost_equal(actual, expected)

    def test_mr_x_mr_index(self):
        cube = CrunchCube(CR.PETS_X_PETS)
        expected = np.array(
            [
                [1.0000000000000000, 1.1724137931034484, 0.894736842105263],
                [0.8529411764705883, 1.0000000000000000, 0.763157894736842],
                [1.1176470588235294, 1.310344827586207, 1.0000000000000000],
            ]
        )
        actual = Index.data(cube, weighted=True, prune=False)
        np.testing.assert_almost_equal(actual, expected)

    def test_ca_items_x_ca_cat_x_cat_index(self):
        cube = CrunchCube(CR.CA_ITEMS_X_CA_CAT_X_CAT)

        # Test index values
        expected = np.array(
            [
                [
                    [1.0, 1.0, np.nan, np.nan, np.nan],
                    [np.nan, np.nan, np.nan, np.nan, np.nan],
                    [np.nan, np.nan, np.nan, np.nan, np.nan],
                    [np.nan, np.nan, np.nan, np.nan, np.nan],
                ],
                [
                    [2.0, 0.0, np.nan, np.nan, np.nan],
                    [0.0, 2.0, np.nan, np.nan, np.nan],
                    [np.nan, np.nan, np.nan, np.nan, np.nan],
                    [np.nan, np.nan, np.nan, np.nan, np.nan],
                ],
                [
                    [np.nan, np.nan, np.nan, np.nan, np.nan],
                    [2.0, 0.0, np.nan, np.nan, np.nan],
                    [0.0, 2.0, np.nan, np.nan, np.nan],
                    [np.nan, np.nan, np.nan, np.nan, np.nan],
                ],
            ]
        )
        actual = Index.data(cube, weighted=True, prune=False)
        np.testing.assert_almost_equal(actual, expected)

        # Test pruning mask
        expected = np.array(
            [
                [
                    [False, False, True, True, True],
                    [True, True, True, True, True],
                    [True, True, True, True, True],
                    [True, True, True, True, True],
                ],
                [
                    [False, False, True, True, True],
                    [False, False, True, True, True],
                    [True, True, True, True, True],
                    [True, True, True, True, True],
                ],
                [
                    [True, True, True, True, True],
                    [False, False, True, True, True],
                    [False, False, True, True, True],
                    [True, True, True, True, True],
                ],
            ]
        )
        actual = Index.data(cube, weighted=True, prune=True).mask
        np.testing.assert_array_equal(actual, expected)

    def test_econ_x_ideology_index_by_col(self):
        cube = CrunchCube(CR.ECON_GENDER_X_IDEOLOGY_WEIGHTED)
        expected = np.array(
            [
                [0.91861761, 0.96399471, 0.9394101, 1.06629132, 1.30027051, 0.67660435],
                [1.08138239, 1.03600529, 1.0605899, 0.93370868, 0.69972949, 1.32339565],
            ]
        )
        actual = Index.data(cube, weighted=True, prune=False)
        np.testing.assert_almost_equal(actual, expected)

    def test_econ_x_ideology_index_by_row(self):
        cube = CrunchCube(CR.ECON_GENDER_X_IDEOLOGY_WEIGHTED)
        expected = np.array(
            [
                [0.91861761, 0.96399471, 0.9394101, 1.06629132, 1.30027051, 0.67660435],
                [1.08138239, 1.03600529, 1.0605899, 0.93370868, 0.69972949, 1.32339565],
            ]
        )
        actual = Index.data(cube, weighted=True, prune=False)
        np.testing.assert_almost_equal(actual, expected)

    def test_cat_x_cat_index_by_col_prune_cols(self):
        cube = CrunchCube(CR.CAT_X_CAT_WITH_EMPTY_COLS)
        expected = np.array(
            [
                [2.8, 0.7, np.nan, 0.7],
                [np.nan, np.nan, np.nan, np.nan],
                [0.0, 0.58333333, np.nan, 2.33333333],
                [0.0, 1.75, np.nan, 0.0],
                [0.0, 1.16666667, np.nan, 1.16666667],
                [0.0, 1.75, np.nan, 0.0],
            ]
        )
        actual = Index.data(cube, weighted=True, prune=False)
        # Assert index without pruning
        np.testing.assert_almost_equal(actual, expected)

        expected = np.array(
            [
                [2.8, 0.7, 0.7],
                [0.0, 0.58333333, 2.33333333],
                [0.0, 1.75, 0.0],
                [0.0, 1.16666667, 1.16666667],
                [0.0, 1.75, 0.0],
            ]
        )
        table = Index.data(cube, weighted=True, prune=True)
        # Assert index witih pruning
        actual = table[:, ~table.mask.all(axis=0)][~table.mask.all(axis=1), :]
        np.testing.assert_almost_equal(actual, expected)

        pruned_expected = [
            np.array([False, True, False, False, False, False]),
            np.array([False, False, True, False]),
        ]
        pruned = cube.prune_indices()
        self.assertEqual(len(pruned), len(pruned_expected))
        for i, actual in enumerate(pruned):
            np.testing.assert_array_equal(actual, pruned_expected[i])



# encoding: utf-8

"""MinBaseSize class."""

from __future__ import division

import numpy as np

from cr.cube.util import lazyproperty
from cr.cube.enums import DIMENSION_TYPE as DT


class MinBaseSizeMask(object):
    """Helper for deciding which rows/columns to suppress, based on min base size.

    If a base value, that is used when calculating percentages, is less than a given
    minimum base size, then all of the values obtained in such a way need to
    suppressed. We achieve this by generating a mask, based on row/column/table
    marginal values and the shape of the underlying slice.
    """

    def __init__(self, slice_, size, hs_dims=None, prune=False):
        self._slice = slice_
        self._size = size
        self._hs_dims = hs_dims
        self._prune = prune

    @lazyproperty
    def column_mask(self):
        """ndarray, True where column margin <= min_base_size, same shape as slice."""
        mask = self._slice.column_base < self._size

        if self._slice.column_base.shape == self._slice.shape:
            # If margin shape is the same as slice's (such as in a col margin for
            # MR x CAT), don't broadcast the mask to the array shape, since
            # they're already the same.
            return mask

        # If the row margin is a row vector - broadcast it's mask to the array shape
        return np.logical_or(np.zeros(self._slice.shape, dtype=bool), mask)

    @lazyproperty
    def row_mask(self):
        """ndarray, True where row margin <= min_base_size, same shape as slice."""
        mask = self._slice.row_base < self._size

        if self._slice.row_base.shape == self._slice.shape:
            # If margin shape is the same as slice's (such as in a row margin for
            # CAT x MR), don't broadcast the mask to the array shape, since
            # they're already the same.
            return mask

        # If the row margin is a column vector - broadcast it's mask to the array shape
        return np.logical_or(np.zeros(self._slice.shape, dtype=bool), mask[:, None])

    @lazyproperty
    def table_mask(self):
        """ndarray, True where table margin <= min_base_size, same shape as slice."""
        mask = self._slice.table_base < self._size

        if self._slice.table_base.shape == self._slice.shape:
            return mask

        if self._slice.dimension_types[0] == DT.MR:
            # If the margin is a column vector - broadcast it's mask to the array shape
            return np.logical_or(np.zeros(self._slice.shape, dtype=bool), mask[:, None])

        return np.logical_or(np.zeros(self._slice.shape, dtype=bool), mask)
